﻿#light

module MSRC.Graph

open System
open System.IO
open System.Collections.Generic
open System.Diagnostics
open MSRC.Inference
open MSRC.Inference.Collections
open MSRC.Inference.Distributions
open MSRC.Inference.Distributions.Gaussians
open MSRC.Inference.Schedule
open MSRC.Dataset
open MSRC.Inference
open Microsoft.FSharp.Collections.ResizeArray
open Microsoft.FSharp.Collections.Array2
open MSRC.Inference.Factors

#if linux
open System.Runtime.InteropServices

[<DllImport("get_time.so", EntryPoint="get_time")>]
let get_time() : double = 0.0
#endif

type Workspace <[<Measure>] 'a> (modelParam : ModelParameter<'a>, metaData : MetaData, 
                                    debugInfo:DebugParam, typeInfo:string) = 

    let noClass = metaData.NoLabel
    let _varBag = DistributionBag<Gaussian<'a>> (Gaussian<'a>.Create (modelParam.weightPriorMu, modelParam.weightPriorSigma))    
    let _weightVar = Array2.init noClass metaData.NoFeature (fun label fea -> _varBag.NewDistribution ())
    
    let (_thresholdVar, _discriminantFactors, _noisyDiscriminantFactors, _differenceFactors, _lossFactors, _GF, 
            _runningVarBag, _runningFactorList) = 
            
        if typeInfo = "ep" then
            (Some(_varBag.NewDistribution(Gaussian.Create (modelParam.thresholdPriorMu, modelParam.thresholdPriorSigma))),
             Some(Array.zero_create<IFactor<'a>> noClass),
             Some(Array.zero_create<IFactor<'a>> noClass),
             Some(Array.zero_create<IFactor<'a>> noClass),
             Some(Array.zero_create<IFactor<'a>> noClass), 
             Some(new GaussianFactors<'a> ()),
             Some(DistributionBag<_> (Gaussian<'a> (0.0<_>, 0.0<_>))),
             Some(FactorList ())
            )
        else (None, None, None, None, None, None, None, None)

    let totalEPIter = ref 0

    do
        if modelParam.useBias then
            let biasIndex = metaData.FeatureNameToId.["b"]
            let precision = (1.0 / (modelParam.biasPriorSigma * modelParam.biasPriorSigma)) * 1.0<_>
            [0..(metaData.NoLabel-1)] |> Seq.iter (fun label ->
                    _varBag.[_weightVar.[label, biasIndex]] <- new Gaussian<'a> (0.0<_>, precision))

        if typeInfo = "ep" then (Option.get _GF).ClearMessages ()
        printfn "Weight allocation successful."

    member this.varBag with get () : DistributionBag<Gaussian<'a>> = _varBag
    member this.weightVar with get () = _weightVar
    
    member this.thresholdVar with get () =  if typeInfo = "ep" then Option.get _thresholdVar 
                                            else failwith "No thresholdVar since EP is not used"
                                            
    member this.discriminantFactors with get () : array<IFactor<'a>> = 
                                            if typeInfo = "ep" then Option.get _discriminantFactors
                                            else failwith "No discriminantFactors since EP is not used"
    
    member this.noisyDiscriminantFactors with get () : array<IFactor<'a>> = 
                                            if typeInfo = "ep" then Option.get _noisyDiscriminantFactors
                                            else failwith "No noisyDiscriminantFactors since EP is not used"
                                            
    member this.differenceFactors with get () : array<IFactor<'a>> = 
                                            if typeInfo = "ep" then Option.get _differenceFactors
                                            else failwith "No differenceFactors since EP is not used"
        
    member this.lossFactors with get () : array<IFactor<'a>> = 
                                            if typeInfo = "ep" then Option.get _lossFactors
                                            else failwith "No lossFactors since EP is not used"
                                            
    member this.runningVarBag with get () : DistributionBag<_> = 
                                            if typeInfo = "ep" then Option.get _runningVarBag
                                            else failwith "No runningVarBag since EP is not used"
                                            
    member this.runningFactorList with get () : FactorList<'a> = 
                                            if typeInfo = "ep" then Option.get _runningFactorList
                                            else failwith "No runningFactorList since EP is not used"
                                            
    member this.GF with get () : GaussianFactors<'a> = 
                                            if typeInfo = "ep" then Option.get _GF
                                            else failwith "No GF since EP is not used"
        
    member this.EPIter with get () = !totalEPIter and set (value) = totalEPIter := value
    member this.getType with get () = typeInfo
    
    member this.showModel () = 
    
        if typeInfo = "ep" then            
            let threshVar_ = Option.get _thresholdVar
            printfn "Global threshold:\t%g\t%g" (float _varBag.[threshVar_].Mu) (float _varBag.[threshVar_].Sigma)
        
        [0..(metaData.NoLabel - 1)] |> Seq.iter (fun label -> 
                [0..(metaData.NoFeature - 1)] |> Seq.iter (fun feaId ->
                        let gau = _varBag.[_weightVar.[label, feaId]]
                        let show = (Int32.to_string(label) + "\t" + metaData.FeatureIdToName.[feaId] + "\t"
                                    + (Float.to_string (gau.Mu * 1.0<_>)) + "\t" + (Float.to_string (gau.Sigma * 1.0<_>)))
                        printfn "%s" show
                    )
                )
        printfn "--------------------------------------------------------------------"

    member this.recordModel () =

        let filename = sprintf @"%s/model_%s.dat" debugInfo.debugDumpDir (modelParam.getIDString())
        let writer = new StreamWriter (filename)
        
        if typeInfo = "ep" then            
            let threshVar_ = Option.get _thresholdVar
            let (mu, sigma) = (_varBag.[threshVar_].Mu, _varBag.[threshVar_].Sigma)
            writer.WriteLine ("Threshold:\t" + (Float.to_string (mu * 1.0<_>)) + "\t" + (Float.to_string (sigma * 1.0<_>)))

        [0..(metaData.NoLabel - 1)] |> Seq.iter (fun label -> 
                [0..(metaData.NoFeature - 1)] |> Seq.iter (fun feaId ->
                        let gau = _varBag.[_weightVar.[label, feaId]]
                        writer.WriteLine (Int32.to_string(label) + "\t" + metaData.FeatureIdToName.[feaId] + "\t"
                                    + (Float.to_string (gau.Mu * 1.0<_>)) + "\t" + (Float.to_string (gau.Sigma * 1.0<_>)))
                    )
                )
        writer.Close()
        printfn "==========  Model recorded in %s ==========" filename


    /// Give an overview of the weights, including max and min of the tau and pi over all the classes
    member this.print_weight_overview () =
    
        let noClass = metaData.NoLabel
        let noFeature = metaData.NoFeature
        
        let mutable cur_min_tau = Double.MaxValue
        let mutable cur_max_tau = Double.MinValue
        let mutable cur_min_pi = Double.MaxValue
        let mutable cur_max_pi = Double.MinValue
        
        for i in 0..(noClass-1) do
            for j in 0..(noFeature - 1) do
                let gau = _varBag.[_weightVar.[i,j]]
                printfn "%d,%d = (%A,%A)" i j gau.Mean gau.Variance
                if gau.Pm < cur_min_tau * 1.0<_> then
                    cur_min_tau <- gau.Pm * 1.0<_>
                if gau.Pm > cur_max_tau * 1.0<_> then
                    cur_max_tau <- gau.Pm * 1.0<_>

                if gau.Pre < cur_min_pi * 1.0<_> then
                    cur_min_pi <- gau.Pre * 1.0<_>
                if gau.Pre > cur_max_pi * 1.0<_> then
                    cur_max_pi <- gau.Pre * 1.0<_>

                if Double.IsNaN (gau.Pre*1.0<_>)|| Double.IsNaN (gau.Pm*1.0<_>) then
                    failwith "oops"
        
        if typeInfo = "ep" then
            let thre = _varBag.[Option.get _thresholdVar]
            printfn "tau %g ~ %g, pi %g ~ %g, thre = (%f,%f)" 
                    cur_min_tau cur_max_tau cur_min_pi cur_max_pi (thre.Pm*1.0<_>) (thre.Pre*1.0<_>)
        else
            printfn "tau %g ~ %g, pi %g ~ %g" 
                    cur_min_tau cur_max_tau cur_min_pi cur_max_pi


type GraphicalModelADF<[<Measure>] 'a> (learningTask:LearningTask<'a>) =

    // Compute the scores of the prediction given the true labels
    // (precision, recall, F1)

    let (trainData, validData, testData, metaData, modelParam, debugInfo) = 
                (learningTask.dataBundle.trainData, learningTask.dataBundle.validData, learningTask.dataBundle.testData, 
                 learningTask.dataBundle.metaData, learningTask.modelParam, learningTask.debugInfo)

    let workspace_ep = if modelParam.useEP then Some(new Workspace <'a> (modelParam, metaData, debugInfo, "ep")) else None
    let workspace_closed = if modelParam.useClosed then 
                                Some(new Workspace <'a> (modelParam, metaData, debugInfo, "closed")) 
                           else None

    let getModel (flag:string) =

        let workspace = 
            if flag = "closed" then Option.get workspace_closed
            elif flag = "ep" then Option.get workspace_ep
            else    // EP weights have higher precedence (if it is used)
                if modelParam.useEP then Option.get workspace_ep
                else Option.get workspace_closed
        
        (workspace.varBag, workspace.weightVar, if workspace.getType = "ep" then workspace.thresholdVar else -1)

    let myFileOpenWrite (filename:string) =
        File.Open(filename, FileMode.Create, FileAccess.Write, FileShare.None);

    let saveModel () = 
          if modelParam.useEP then (Option.get workspace_ep).recordModel ()
          if modelParam.useClosed then (Option.get workspace_closed).recordModel ()

    do
        if not modelParam.useEP && not modelParam.useClosed then
            failwith "Oops, neither EP nor Closed update is used."
        
        let record_poterior = null
        GC.Collect ()        
            
    /// Prototype: (trueLabels : array<int>, predictLabel : ResizeArray<int>) -> (precision, recall, Fscore)
    /// Func: For one document, given the true labels and prediced labels, compute the (precision, recall, Fscore)
    /// No longer in use.
    member this.getScores (trueLabels : array<int>, predictLabel : ResizeArray<int>) =
    
        let minLength = min trueLabels.Length predictLabel.Count
        
        let mutable true_pos = 0
        let mutable idx1 = 0
        let mutable idx2 = 0

        while idx1 < minLength && idx2 < minLength do
            if trueLabels.[idx1] < predictLabel.[idx2] then
                idx1 <- idx1 + 1
            elif trueLabels.[idx1] > predictLabel.[idx2] then
                idx2 <- idx2 + 1
            else
                true_pos <- true_pos + 1   
                idx1 <- idx1 + 1     
                idx2 <- idx2 + 1
        
        let precision = (float true_pos) / (float predictLabel.Count) * 100.0
        let recall = (float true_pos) / (float trueLabels.Length) * 100.0
        if recall |> System.Double.IsNaN then
            failwith "trueLabels.Length = %d" trueLabels.Length
            
        if predictLabel.Count = 0 then
            (0.0, recall, 0.0)
        else
            let Fscore = if recall > 1e-5 && precision > 1e-5 then
                            2.0 * recall * precision / (recall + precision)
                         else 0.0
            (precision, recall, Fscore)

            
    /// Returns (acc_total_pos, acc_true_pos), both are (numPredThreshold * noClass) Array2
    /// total_pos := true_pos + false_pos
    member this.get_cumulation_on_thresholds (predThreshGrid:seq<float>, proofData:SeqDataSet<'a>, 
                                                metaData:MetaData, varBag:DistributionBag<Gaussian<'a>>, 
                                                weightVar:int[,], thresholdVar:int) =

        let noClass = metaData.NoLabel
        let numPredThreshold = Seq.length predThreshGrid

        let acc_total_pos = Array2.create numPredThreshold noClass 0
        let acc_true_pos = Array2.create numPredThreshold noClass 0         

        let rec getMicroMacroFScore (proofdata:SeqDataSet<'a>) =

            match proofdata.getNextInstance () with
            | None -> ()
            | Some(instance) -> 
                let curActiveLabels = instance.Labels
                let nextId = ref 0
                [0 .. (noClass - 1)] |> Seq.iter (fun label ->
                    let (disc_mu, disc_sigma2) = instance.Features |> Array.fold_left  (fun (acc_mu, acc_sigma2) (feaId, value) ->   
                                                    let gau = varBag.[weightVar.[label, feaId]]                                                
                                                    (acc_mu + value * gau.Mu, acc_sigma2 + value * value * gau.Variance ))
                                                    (-varBag.[thresholdVar].Mu, varBag.[thresholdVar].Variance)
                    let cdf = normcdf(disc_mu / sqrt(disc_sigma2))

                    let label_relevant = !nextId < instance.NoLabel && curActiveLabels.[!nextId] = label
                    if label_relevant then nextId := !nextId + 1

                    predThreshGrid |> Seq.iteri (fun thresh_idx threshold ->
                        if cdf > threshold then
                            acc_total_pos.[thresh_idx, label] <- acc_total_pos.[thresh_idx, label] + 1
                            if label_relevant then
                                acc_true_pos.[thresh_idx, label] <- acc_true_pos.[thresh_idx, label] + 1
                    )
                )
                curActiveLabels |> Array.iter (fun label -> 
                                    predThreshGrid |> Seq.iteri (fun thresh_idx threshold ->
                                    acc_total_pos.[thresh_idx, label] <- acc_total_pos.[thresh_idx, label] + 1)) 

                getMicroMacroFScore proofdata

        proofData.rewind ()
        getMicroMacroFScore proofData
        
        (acc_total_pos, acc_true_pos)


        
        
    /// Tune the probability threshold of each class on the validation dataset
    /// Returns 
    ///  1: an array of threshold, each component is the optimal threshold for a class (for macroaverage)
    ///  2: two scalar values as common threshold for all classes, which optimize the microaverage and macroaverage
    member this.tuneThreshold () =

        let noClass = metaData.NoLabel
        let threshold = Array.zero_create<float> noClass

        let (varBag, weightVar, thresholdVar) = getModel ("")
        
        let candidateThreshold = [| 0.001; 0.01; 0.04; 0.07; 0.1; 0.15; 0.2; 0.25; 0.3; 0.35; 0.4; 0.45 |]
        let numPredThreshold = Seq.length candidateThreshold
        
        // First get the cumulant information on the validation data using various thresholds
        printfn "Tuning threshold with validation data."                
        let (acc_total_pos, acc_true_pos) = 
                this.get_cumulation_on_thresholds (candidateThreshold, validData, metaData, varBag, weightVar, thresholdVar)
        
               
        
        // Compute the final F-score performance under various thresholds
        // Find the argmax of threshold for each class
        [0..(noClass - 1)] |> Seq.iter (fun label ->
                let (maxScore, argMax) = 
                    [| 0..(numPredThreshold-1) |] |> Array.fold_left (fun (maxVal, maxIdx) threshIdx ->
                            if acc_total_pos.[threshIdx,label] = 0 then
                                (maxVal, maxIdx)
                            else
                                let temp = (float acc_true_pos.[threshIdx,label]) / (float acc_total_pos.[threshIdx,label])
                                if temp > maxVal then
                                    (temp, threshIdx)
                                else
                                    (maxVal, maxIdx) )
                            (Double.MinValue, 0)
                threshold.[label] <- candidateThreshold.[argMax])
                

        // Second: find a common threshold which maximizes the microaverage Fscore
        let (maxValidScore_micro, argmaxValidScore_micro) = 
                [| 0..(numPredThreshold-1) |] |> Array.fold_left (fun (curMax, argmax) threshIdx ->
                                let (totalTruePos, totalTotalPos) =              
                                    [|0..(noClass-1)|] |> Array.fold_left (fun (acc1, acc2) label ->                            
                                        if acc_total_pos.[threshIdx, label] > 0 then
                                            ( acc1 + acc_true_pos.[threshIdx, label], 
                                              acc2 + acc_total_pos.[threshIdx, label] )
                                        else
                                            (acc1, acc2)
                                    ) (0, 0)
                                    
                                if (float totalTruePos) / (float totalTotalPos) > curMax then
                                    ((float totalTruePos) / (float totalTotalPos), threshIdx)
                                else
                                    (curMax, argmax)
                               ) (Double.MinValue, 0)


        // Third: find a common threshold which maximizes the macroaverage Fscore
        let (maxValidScore_macro, argmaxValidScore_macro) = 
                [| 0..(numPredThreshold-1) |] |> Array.fold_left (fun (curMax, argmax) threshIdx ->
                                
                                let (numEffectClass, totalFscore) = 
                                    [|0..(noClass-1)|] |> Array.fold_left (fun (acc1, acc2) label -> 
                                        if acc_total_pos.[threshIdx, label] > 0 then
                                            (acc1 + 1,
                                             acc2 + (float acc_true_pos.[threshIdx, label]) / (float acc_total_pos.[threshIdx, label])
                                             )
                                        else
                                            (acc1, acc2)
                                       ) (0, 0.0)

                                if totalFscore / (float numEffectClass) > curMax then
                                    (totalFscore / (float numEffectClass), threshIdx)
                                else
                                    (curMax, argmax)
                               ) (Double.MinValue, 0)
                               
                       
        (threshold, candidateThreshold.[argmaxValidScore_micro], candidateThreshold.[argmaxValidScore_macro])
           
        
    /// Enumerate the common thresholds for the micro-averaged and macro-averaged F1 score (shared by all categories)
    /// And display the resulting micro-averaged and macro-averaged F1 score
    member this.enumMicroMacroFscore (testData:SeqDataSet<'a>, metaData:MetaData, varBag:DistributionBag<Gaussian<'a>>, 
                                        weightVar:int[,], thresholdVar:int) = 
            
        let noClass = metaData.NoLabel

//            this.recordModel (varBag, weightVar, thresholdVar, modelDir, metaData)

//            let logger = new StreamWriter("show_discriminant.txt")
        let predThreshGrid = [ 0.001; 0.01; 0.04; 0.07; 0.1; 0.15; 0.2; 0.25; 0.3; 0.35; 0.4; 0.45]
        
        // First get the cumulant information on the validation data using various thresholds
        printfn "Testing with thresholds enumerated."        
        let (acc_total_pos, acc_true_pos) = 
                this.get_cumulation_on_thresholds (predThreshGrid, testData, metaData, varBag, weightVar, thresholdVar)
        
        [0..(predThreshGrid.Length-1)] |> Seq.iter (fun thresh_idx ->
        
            let (totalTruePos, totalTotalPos, numEffectiveClass, macroaverage_baby) =
                        [|0..(noClass-1)|] |> Array.fold_left (fun (acc1, acc2, acc3, acc4) label ->
                            if acc_total_pos.[thresh_idx, label] > 0 then
                                ( acc1 + acc_true_pos.[thresh_idx, label], 
                                  acc2 + acc_total_pos.[thresh_idx, label],
                                  acc3 + 1,
                                  acc4 + (float acc_true_pos.[thresh_idx, label]) / (float acc_total_pos.[thresh_idx, label]))
                            else
                                (acc1, acc2, acc3, acc4)
                        ) (0, 0, 0, 0.0)

            let macroaverage = macroaverage_baby * 200.0 / (float numEffectiveClass)
            let microaverage = 200.0 * (float totalTruePos) / (float totalTotalPos)

            printfn "predThreshold = %g, numEffectiveClass = %d, Macroaverage Fscore = ^^ %g ^^, Microaverage Fscore = ^^ %g ^^" 
                    predThreshGrid.[thresh_idx] numEffectiveClass macroaverage microaverage                
        )


    /// Use the given probability threshold to calculate the micro-averaged and macro-averaged F1 score
    /// Input 
    ///   1. An array of probability threshold, one for each class
    ///   2. Two common thresholds to be shared by all classes (derived by maximizing microaverage and macroaverage respectively)
    /// Has been replaced by fastTestMicroMacroFscore.
    member this.testMicroMacroFscore (probThresh:array<float>, commonThresh_micro:float, commonThresh_macro:float) = 

        let noClass = metaData.NoLabel

        let (varBag, weightVar, thresholdVar) = getModel ("")

//      this.recordModel (varBag, weightVar, thresholdVar, modelDir, metaData)

//      let logger = new StreamWriter("show_discriminant.txt")
                
        let (acc_total_pos, acc_true_pos) = (Array.create noClass 0, Array.create noClass 0)
        
        let (acc_total_pos_common_macro, acc_true_pos_common_macro, acc_total_pos_common_micro, acc_true_pos_common_micro) =             
                (Array.create noClass 0, Array.create noClass 0, 
                    Array.create noClass 0, Array.create noClass 0)
                
        printfn "Predicting on testing data."

        let rec getCumulant (testdata:SeqDataSet<'a>) =
            match testdata.getNextInstance () with
            | None -> ()
            | Some(instance) -> 
                let curActiveLabels = instance.Labels
                let nextId = ref 0
                [0 .. (noClass - 1)] |> Seq.iter (fun label ->
                    let (disc_mu, disc_sigma2) = instance.Features |> Array.fold_left  (fun (acc_mu, acc_sigma2) (feaId, value) ->   
                                                    let gau = varBag.[weightVar.[label, feaId]]                                                
                                                    (acc_mu + value * gau.Mu, acc_sigma2 + value * value * gau.Variance ))
                                                    (-varBag.[thresholdVar].Mu, varBag.[thresholdVar].Variance)
                    let cdf = normcdf(disc_mu / sqrt(disc_sigma2))

                    let label_relevant = !nextId < instance.NoLabel && curActiveLabels.[!nextId] = label
                    if label_relevant then nextId := !nextId + 1

                    if cdf >= probThresh.[label] then
                        acc_total_pos.[label] <- acc_total_pos.[label] + 1
                        if label_relevant then
                            acc_true_pos.[label] <- acc_true_pos.[label] + 1
                    
                    if cdf >= commonThresh_macro then
                        acc_total_pos_common_macro.[label] <- acc_total_pos_common_macro.[label] + 1
                        if label_relevant then
                            acc_true_pos_common_macro.[label] <- acc_true_pos_common_macro.[label] + 1

                    if cdf >= commonThresh_micro then
                        acc_total_pos_common_micro.[label] <- acc_total_pos_common_micro.[label] + 1
                        if label_relevant then
                            acc_true_pos_common_micro.[label] <- acc_true_pos_common_micro.[label] + 1
                            
                )
                curActiveLabels |> Array.iter (fun label -> 
                                    acc_total_pos.[label] <- acc_total_pos.[label] + 1
                                    acc_total_pos_common_macro.[label] <- acc_total_pos_common_macro.[label] + 1
                                    acc_total_pos_common_micro.[label] <- acc_total_pos_common_micro.[label] + 1
                                    )

                getCumulant testdata

        testData.rewind ()
        getCumulant testData
       
        let (totalTruePos, totalTotalPos, numEffectiveClass, macroaverage_baby,
             totalTruePos_common_macro, totalTotalPos_common_macro, numEffectiveClass_common_macro, macroaverage_baby_common_macro,
             totalTruePos_common_micro, totalTotalPos_common_micro, numEffectiveClass_common_micro, macroaverage_baby_common_micro) =
            [|0..(noClass-1)|] |> Array.fold_left (fun (acc1, acc2, acc3, acc4, 
                                                        acc1_common_macro, acc2_common_macro, acc3_common_macro, acc4_common_macro,
                                                        acc1_common_micro, acc2_common_micro, acc3_common_micro, acc4_common_micro) label ->
                            let (part1_1, part1_2, part1_3, part1_4) = 
                                if acc_total_pos.[label] > 0 then
                                    ( acc1 + acc_true_pos.[label], 
                                      acc2 + acc_total_pos.[label],
                                      acc3 + 1,
                                      acc4 + (float acc_true_pos.[label]) / (float acc_total_pos.[label]))
                                else
                                    (acc1, acc2, acc3, acc4)
                            let (part2_1, part2_2, part2_3, part2_4) = 
                                if acc_total_pos_common_macro.[label] > 0 then
                                    ( acc1_common_macro + acc_true_pos_common_macro.[label], 
                                      acc2_common_macro + acc_total_pos_common_macro.[label],
                                      acc3_common_macro + 1,
                                      acc4_common_macro + (float acc_true_pos_common_macro.[label]) / (float acc_total_pos_common_macro.[label]))
                                else
                                    (acc1_common_macro, acc2_common_macro, acc3_common_macro, acc4_common_macro)
                            let (part3_1, part3_2, part3_3, part3_4) = 
                                if acc_total_pos_common_macro.[label] > 0 then
                                    ( acc1_common_micro + acc_true_pos_common_micro.[label], 
                                      acc2_common_micro + acc_total_pos_common_micro.[label],
                                      acc3_common_micro + 1,
                                      acc4_common_micro + (float acc_true_pos_common_micro.[label]) / (float acc_total_pos_common_micro.[label]))
                                else
                                    (acc1_common_micro, acc2_common_micro, acc3_common_micro, acc4_common_micro)
                            (part1_1, part1_2, part1_3, part1_4, part2_1, part2_2, part2_3, part2_4, part3_1, part3_2, part3_3, part3_4)
                        ) (0, 0, 0, 0.0, 0, 0, 0, 0.0, 0, 0, 0, 0.0)


        let macroaverage_common_macro = macroaverage_baby_common_macro * 200.0 / (float numEffectiveClass_common_macro)
        let microaverage_common_macro = 200.0 * (float totalTruePos_common_macro) / (float totalTotalPos_common_macro)

        let macroaverage_common_micro = macroaverage_baby_common_micro * 200.0 / (float numEffectiveClass_common_micro)
        let microaverage_common_micro = 200.0 * (float totalTruePos_common_micro) / (float totalTotalPos_common_micro)

        let macroaverage = macroaverage_baby * 200.0 / (float numEffectiveClass)
        let microaverage = 200.0 * (float totalTruePos) / (float totalTotalPos)

        printfn "numEffectiveClass = %d, Macroaverage Fscore = ^^ %g ^^, Microaverage Fscore = ^^ %g ^^"
                   numEffectiveClass macroaverage microaverage                

        printfn "Using common threshold chosen by microaverage = %g, numEffectiveClass = %d, Macroaverage Fscore = ^^ %g ^^, Microaverage Fscore = ^^ %g ^^" 
                commonThresh_macro numEffectiveClass_common_micro macroaverage_common_micro microaverage_common_micro
                
        printfn "Using common threshold chosen by macroaverage = %g, numEffectiveClass = %d, Macroaverage Fscore = ^^ %g ^^, Microaverage Fscore = ^^ %g ^^" 
                        commonThresh_micro numEffectiveClass_common_macro macroaverage_common_macro microaverage_common_macro











    member this.generateSeeds () =
    
        let rnd = System.Random(0)
        let seedArray = Array2.init metaData.NoLabel metaData.NoFeature (fun _ _ ->
                        rnd.Next () )
                        
        (seedArray, rnd.Next ())   



    //////////////////////////////////////////////////////////////////////
    /// Now the beast
    /// Implementation of the sampling posterior to find the threshold
    /// Returns the best threshold found for each category
    /// 4 heuristics are tried in case the sampled weights never output positive for a certain class, 
    /// Also tried the halving trick
    /// In sum, we have smallestTrial+argmax, smallestTrial+argmax/2, greatestTrial+argmax, greatestTrial+argmax/2 
    ///                 smallestProb+argmax, smallestProb+argmax/2, greatestProb+argmax, greatestProb+argmax/2

    member this.correlationDiagnosis_singleRead (testData:SeqDataSet<'a>, metaData:MetaData, betaSquare:float<'a^2>,
                                                    varBag:DistributionBag<Gaussian<'a>>, weightVar:int[,], 
                                                    thresholdVar:int, trainIter:int) = 

        let noClass = metaData.NoLabel
        
        let threshGrid = [| 0.000001; 0.01; 0.05; 0.1; 0.15; 0.2; 0.25; 0.3; 0.35; 0.4; 0.5; 0.55; 0.6; 0.65; 0.7; 0.75; 0.8; 1.1 |]
        let noThreshGrid = threshGrid.Length
        let noSample = modelParam.noSample

        let (weightSeed, biasSeed) = this.generateSeeds()         

        let alpha = Array3.zero_create noClass noSample noThreshGrid
        let beta = Array2.zero_create noClass noSample
        let gamma = Array2.zero_create noClass noThreshGrid

        let eta = Array2.zero_create noClass noThreshGrid
        let rho = Array.zero_create noClass

        let ExpFs = Array2.zero_create noClass noThreshGrid

        let minClassThresh = Array.zero_create noClass        
        let maxClassThresh = Array.zero_create noClass

        let thresholds = Array2.zero_create 8 noClass
        let thresholdNames = [| @"smallestTrial+argmax"; @"smallestTrial+argmax/2"; @"greatestTrial+argmax"; @"greatestTrial+argmax/2"; 
                                @"smallestProb+argmax"; @"smallestProb+argmax/2"; @"greatestProb+argmax"; @"greatestProb+argmax/2"|]

        printfn "Performing correlation diagnosis."
        
        let resultThreshold = Array.init 8 (fun i -> (Array.zero_create noClass))

        let biasRng = System.Random (biasSeed)  // random number generator for the bias
        let (mu, std) = (varBag.[thresholdVar].Mu, varBag.[thresholdVar].StandardDeviation)
        let biasSample = Array.init noSample (fun i -> 
                mu + std * norminv ((float (biasRng.Next(modelParam.maxRandValue))) / (float modelParam.maxRandValue))
            )
        
        let betaNoise = sqrt(betaSquare)
        let betaSample = Array2.init noSample noClass (fun sample label -> 
                betaNoise * norminv ((float (biasRng.Next(modelParam.maxRandValue))) / (float modelParam.maxRandValue))
            )

        let show = ref 0
        #if linux
        let startTime = get_time() // System.DateTime.Now
        #else
        let pro = Process.GetCurrentProcess ()
        let startTime = pro.TotalProcessorTime // System.DateTime.Now
        #endif
        
        let rec getCumulant (testdata:SeqDataSet<'a>) =            
            match testdata.getNextInstance () with
            | None -> ()
            | Some(instance) -> 

                 show := !show + 1
                 if !show % 100 = 0 then printf "%d done (%d). " !show 
                 #if linux
                                                (Int32.of_float (get_time() - startTime))
                 #else
                                                (Int32.of_float (pro.TotalProcessorTime - startTime).TotalSeconds)
                                                // (Int32.of_float (System.DateTime.Now - startTime).TotalSeconds)
                 #endif
                 
                 let curActiveLabels = instance.Labels
                 let nextId = ref 0
                 
                 curActiveLabels |> Array.iter (fun label -> rho.[label] <- rho.[label] + 1)

                 [0 .. (noClass - 1)] |> Seq.iter (fun label ->
                    let (disc_mu, disc_sigma2) = instance.Features |> Array.fold_left  (fun (acc_mu, acc_sigma2) (feaId, value) ->
                                                    let gau = varBag.[weightVar.[label, feaId]]
                                                    (acc_mu + value * gau.Mu, acc_sigma2 + value * value * gau.Variance ))
                                                    (-varBag.[thresholdVar].Mu, varBag.[thresholdVar].Variance)

                    let marginProb = normcdf(disc_mu / sqrt(disc_sigma2))
                    if marginProb > maxClassThresh.[label] then maxClassThresh.[label] <- marginProb
                    if marginProb < maxClassThresh.[label] then minClassThresh.[label] <- marginProb

                    let label_relevant = !nextId < instance.NoLabel && curActiveLabels.[!nextId] = label
                    if label_relevant then nextId := !nextId + 1

                    let threshFlag = threshGrid |> Array.mapi (fun threshId thresh -> 
                        if marginProb > thresh then
                            gamma.[label, threshId] <- gamma.[label, threshId] + 1

                            // now check with ground truth
                            if label_relevant then
                                eta.[label, threshId] <- eta.[label, threshId] + 1
                            true
                        else
                            false )

                    if noSample > 0 then
                        // Create random number generator
                        let rngList = instance.Features |> Array.map (fun (feaId, _) ->
                            System.Random (weightSeed.[label, feaId]))

                        [0..(noSample-1)] |> Seq.iter (fun sampleId -> 
                                let sampleY = [| 0..(instance.Features.Length - 1) |] |> Array.fold_left (fun acc nonzeroId ->
                                                  let (feaId, feaVal) = instance.Features.[nonzeroId]
                                                  let gau = varBag.[weightVar.[label, feaId]]

                                                  // draw a random number from N(0,1)
                                                  let randGauNum = norminv ((float (rngList.[nonzeroId].Next(modelParam.maxRandValue))) 
                                                                                / (float modelParam.maxRandValue))
                                                  acc + (gau.StandardDeviation * randGauNum + gau.Mu) * feaVal
                                              )  (-biasSample.[sampleId] + betaSample.[sampleId, label])

                                if sampleY > 0.0<_> then
                                    beta.[label, sampleId] <- beta.[label, sampleId] + 1
                                    threshFlag |> Array.iteri (fun threshIdx flag ->
                                        if flag then alpha.[label, sampleId, threshIdx] <- alpha.[label, sampleId, threshIdx] + 1)
                        )                 
                 )

                 getCumulant testdata

        testData.rewind ()
        getCumulant testData


        printfn "\nstatistics collected, costing %f seconds" 
        #if linux
                (get_time() - startTime)
        #else
                (pro.TotalProcessorTime - startTime).TotalSeconds
                // (System.DateTime.Now - startTime).TotalSeconds
        #endif
        
        printfn "--------------------------------------------------------------------"

        // collect the cumulants
        if noSample > 0 then
            // Finally output the result stored in ExpFs.  Loop over all classes.
            [| 0..(noClass-1) |] |> Array.iter (fun label ->
                               let (maxVal, argMax) = 
                                       [| 0..(noThreshGrid-1) |] |> Array.fold_left (fun (curMax, curArgMax) threshIdx ->
                                                   let noEffect = ref 0
                                                   let temp = [0..(noSample-1)] |> Seq.fold (fun acc sampleId ->
                                                               if beta.[label, sampleId] + gamma.[label, threshIdx] > 0 then
                                                                    noEffect := !noEffect + 1
                                                                    acc + (float alpha.[label, sampleId, threshIdx]) / 
                                                                       (float (beta.[label, sampleId] + gamma.[label, threshIdx]))
                                                               else
                                                                    acc
                                                               ) 0.0
                                                   ExpFs.[label, threshIdx] <- temp / (float !noEffect)

                                                   if ExpFs.[label, threshIdx] > curMax then
                                                       (ExpFs.[label, threshIdx], threshGrid.[threshIdx])
                                                   else
                                                       (curMax, curArgMax)
                                                   ) (Double.MinValue, -1.0)

                               if Double.IsNaN ExpFs.[label, noThreshGrid-1] then
//                                    printfn "%d special" label
                                    [0; 1] |> Seq.iter (fun i -> thresholds.[i, label] <- threshGrid.[0])
                                    [2; 3] |> Seq.iter (fun i -> thresholds.[i, label] <- threshGrid.[noThreshGrid-1])
                                    [4; 5] |> Seq.iter (fun i -> thresholds.[i, label] <- minClassThresh.[label])
                                    [6; 7] |> Seq.iter (fun i -> thresholds.[i, label] <- maxClassThresh.[label])                                    
                               else
//                                    printfn "%d choose %g" label argMax
                                    [0; 2; 4; 6] |> Seq.iter (fun i -> thresholds.[i, label] <- argMax)
                                    [1; 3; 5; 7] |> Seq.iter (fun i -> thresholds.[i, label] <- argMax / 2.0)                                            
                            )

        // Write to file the empiracal estimation of expected F-score
        if modelParam.writeSampleExpFs then  
                         
            let writer = new StreamWriter(myFileOpenWrite(sprintf "sound_check_%s_%s_sample=%d_prior=%A_noise=%g.dat"
                                                            modelParam.datasetName modelParam.filterFname modelParam.noSample 
                                                            modelParam.biasPriorSigma (betaNoise * 1.0<_>))) //
            threshGrid |> Array.iter (fun value ->
                writer.Write(Float.to_string (value) + "\t"))
            writer.Write("\n")

            [| 0..(noClass-1) |] |> Array.iter (fun label ->
                                        [| 0..(noThreshGrid-1) |] |> Array.iter (fun idx ->
                                            writer.Write( sprintf "%.5g\t" (ExpFs.[label, idx] * 200.0)))
                                        writer.Write("\n")
                                    )
            writer.Close ()


        // Write to file the F-score(l(\theta), l^*)
        if modelParam.writeRefFscore then
        
            let writer = new StreamWriter(myFileOpenWrite(sprintf "ground_truth_%s_%s_sample=%d_prior=%A_noise=%g.dat"
                                                            modelParam.datasetName modelParam.filterFname modelParam.noSample 
                                                            modelParam.biasPriorSigma (betaNoise * 1.0<_>)))  
            threshGrid |> Array.iter (fun value ->
                writer.Write(Float.to_string (value) + "\t"))
            writer.Write("\n")
        
            [| 0..(noClass-1) |] |> Array.iter (fun label ->
                                    [| 0..(noThreshGrid-1) |] |> Array.iter (fun threshIdx ->
                                        let value = (float eta.[label, threshIdx]) / (float (gamma.[label, threshIdx] + rho.[label]))
                                        writer.Write( (sprintf "%.5g\t" (value * 200.0))))
                                    writer.Write("\n")
                                )
            writer.Close ()

        (thresholds, thresholdNames)
        


    member this.cmd_threshold (testData:SeqDataSet<'a>, metaData:MetaData, betaSquare:float<'a^2>,
                                                    varBag:DistributionBag<Gaussian<'a>>, weightVar:int[,], 
                                                    thresholdVar:int, trainIter:int,
                                                    num_train_instance:int[], num_active) = 
        let noClass = metaData.NoLabel
        let noFeature = metaData.NoFeature
        
        let thresholds = Array2.zero_create 1 noClass
        let thresholdNames = [| @"cmn" |]
    
        let (gbias_mu, gbias_var, gbias_std) = 
                if modelParam.useGlobalBias 
                    then (varBag.[thresholdVar].Mu, varBag.[thresholdVar].Variance, varBag.[thresholdVar].StandardDeviation) 
                else (0.0<_>, 0.0<_>, 0.0<_>)

        let resultThreshold = Array.init 1 (fun i -> (Array.zero_create noClass))

        #if linux
        let startTime = get_time() // System.DateTime.Now
        #else
        let pro = Process.GetCurrentProcess ()
        let startTime = pro.TotalProcessorTime // System.DateTime.Now
        #endif

        let record_posterior = Array.init noClass (fun i -> Array.zero_create testData.NoInstance)
        
        testData.rewind ()            
        let instanceOpt = ref (testData.getNextInstance ())
        let testInstanceID = ref -1
        
        while Option.is_some !instanceOpt do
                        
            let instance = Option.get !instanceOpt
            let curActiveLabels = instance.Labels
            let nextId = ref 0
            testInstanceID := !testInstanceID + 1

            [0 .. (noClass - 1)] |> Seq.iter (fun label ->
                let (disc_mu, disc_sigma2) = instance.Features |> Array.fold_left  (fun (acc_mu, acc_sigma2) (feaId, value) ->
                                                        let gau = varBag.[weightVar.[label, feaId]]
                                                        (acc_mu + value * gau.Mu, acc_sigma2 + value * value * gau.Variance ))
                                                        (-gbias_mu, gbias_var)

                record_posterior.[label].[!testInstanceID] <- disc_mu / sqrt(disc_sigma2 + betaSquare)                           
            )

            instanceOpt := testData.getNextInstance ()

        [0 .. (noClass - 1)] |> Seq.iter (fun label ->
        
                /// sort in decreasing order
                (record_posterior.[label]) |> Array.sort (fun a b -> if a > b then -1 elif a < b then 1 else 0)
                let ratio = (float num_train_instance.[label]) / (float num_active)
                let index =  Int32.of_float ((float testData.NoInstance) * ratio)                
                thresholds.[0, label] <- normcdf(record_posterior.[label].[index])
        )
        
        (thresholds, thresholdNames)


    //////////////////////////////////////////////////////////////////////
    /// Allows reading testing data for multiple times
    /// Implementation of the sampling posterior to find the threshold
    /// Returns the best threshold found for each category
    /// 4 heuristics are tried in case the sampled weights never output positive for a certain class, 
    /// Also tried the halving trick
    /// In sum, we have smallestTrial+argmax, smallestTrial+argmax/2, greatestTrial+argmax, greatestTrial+argmax/2 
    ///                 smallestProb+argmax, smallestProb+argmax/2, greatestProb+argmax, greatestProb+argmax/2

    member this.correlationDiagnosis_multipleRead (testData:SeqDataSet<'a>, metaData:MetaData, betaSquare:float<'a^2>,
                                                    varBag:DistributionBag<Gaussian<'a>>, weightVar:int[,], 
                                                    thresholdVar:int, trainIter:int) = 

        let noClass = metaData.NoLabel
        let noFeature = metaData.NoFeature
        
        // threshGrid must be increasingly sorted
        let threshGrid = [| 0.0001; 0.0005; 0.001; 0.005; 0.01; 0.02; 0.04; 0.06; 0.08; 0.1; 0.15; 0.2; 0.25; 0.3; 0.35; 0.4; 0.5; 0.55; 0.6; 1.1 |]
        let noThreshGrid = threshGrid.Length
        let noSample = modelParam.noSample

        let alpha = Array2.zero_create<int> noClass noThreshGrid
        let beta = Array.zero_create<int> noClass
        let gamma = Array2.zero_create<int> noClass noThreshGrid

        let eta = Array2.zero_create noClass noThreshGrid
        let rho = Array.zero_create noClass

        let ExpFs = Array2.zero_create noClass noThreshGrid
        let noEffective = Array2.zero_create noClass noThreshGrid

        let minClassThresh = Array.zero_create noClass        
        let maxClassThresh = Array.zero_create noClass

        let thresholds = Array2.zero_create 8 noClass
        let thresholdNames = [| @"smallestTrial+argmax"; @"smallestTrial+argmax/2"; @"greatestTrial+argmax"; @"greatestTrial+argmax/2"; 
                                @"smallestProb+argmax"; @"smallestProb+argmax/2"; @"greatestProb+argmax"; @"greatestProb+argmax/2"|]
    
        let (gbias_mu, gbias_var, gbias_std) = 
                if modelParam.useGlobalBias 
                    then (varBag.[thresholdVar].Mu, varBag.[thresholdVar].Variance, varBag.[thresholdVar].StandardDeviation) 
                else (0.0<_>, 0.0<_>, 0.0<_>)
            
        printfn "Performing correlation diagnosis."
        
        let resultThreshold = Array.init 8 (fun i -> (Array.zero_create noClass))

        let rng = System.Random(1)
        
        let biasSample = Array.init noSample (fun i -> 
                gbias_mu + gbias_std * norminv (0.2 + 0.6 * (float (rng.Next(modelParam.maxRandValue))) / (float modelParam.maxRandValue))
            )
            
        let weightSample = Array2.zero_create noClass noFeature

        let betaNoise = sqrt(betaSquare)
        let betaSample = Array2.init noSample noClass (fun sample label -> 
                betaNoise * norminv (0.2 + 0.6 * (float (rng.Next(modelParam.maxRandValue))) / (float modelParam.maxRandValue))
            )

        if not (Directory.Exists (modelParam.root + @"/Dataset/useless")) then
            Directory.CreateDirectory (modelParam.root + @"/Dataset/useless") |> ignore
        
        #if linux
        let startTime = get_time() // System.DateTime.Now
        #else
        let pro = Process.GetCurrentProcess ()
        let startTime = pro.TotalProcessorTime // System.DateTime.Now
        #endif

        let startTimeDate = System.DateTime.Now
        let buffer_filename = modelParam.root + @"/Dataset/useless/threshIdx" + (Int64.to_string startTimeDate.Ticks) + ".dat" 
        let record_posterior = Array.init noClass (fun i -> Array.zero_create testData.NoInstance)

        [0..(noSample-1)] |> Seq.iter (fun sampleId ->
        
            printf "sample %d (%d). " 
            #if linux
                    sampleId (Int32.of_float (get_time() - startTime))
            #else
                    sampleId (Int32.of_float (pro.TotalProcessorTime - startTime).TotalSeconds)
                    // (Int32.of_float (System.DateTime.Now - startTime).TotalSeconds)
            #endif
            
            [0..(noClass-1)] |> Seq.iter (fun label ->             
                    [0..(noFeature-1)] |> Seq.iter (fun feaId -> 
                            let gau = varBag.[weightVar.[label, feaId]]
                            weightSample.[label, feaId] <- gau.StandardDeviation 
                                                            //* norminv ((float (rng.Next(maxRandValue))) / (float maxRandValue))
                                                            * norminv (0.2 + 0.6 * (float (rng.Next(modelParam.maxRandValue))) / (float modelParam.maxRandValue))
                                                            + gau.Mu
                        )
                    beta.[label] <- 0
                    [0..(noThreshGrid-1)] |> Seq.iter (fun threshId -> alpha.[label, threshId] <- 0)
                  )
            
            let threshIdx_reader = if sampleId > 0 then 
                                        new BinaryReader(File.Open(buffer_filename, FileMode.Open, FileAccess.Read, FileShare.None))
                                   else null

            let threshIdx_writer = if sampleId = 0 then 
                                        new BinaryWriter(File.Open(buffer_filename, FileMode.Create, FileAccess.Write, FileShare.None))
                                   else null
        
            testData.rewind ()            
            let instanceOpt = ref (testData.getNextInstance ())
            let testInstanceID = ref -1
            
            while Option.is_some !instanceOpt do
                            
                let instance = Option.get !instanceOpt
                let curActiveLabels = instance.Labels
                let nextId = ref 0
                testInstanceID := !testInstanceID + 1
                
                if sampleId = 0 then        // for computing reference to the ground truth
                    curActiveLabels |> Array.iter (fun label -> rho.[label] <- rho.[label] + 1)
                
                [0 .. (noClass - 1)] |> Seq.iter (fun label ->
                
                    let label_relevant = !nextId < instance.NoLabel && curActiveLabels.[!nextId] = label
                    if label_relevant then nextId := !nextId + 1

                    let threshMaxIdx = 
                        if sampleId = 0 then 
                            let (disc_mu, disc_sigma2) = instance.Features |> Array.fold_left  (fun (acc_mu, acc_sigma2) (feaId, value) ->
                                                            let gau = varBag.[weightVar.[label, feaId]]
                                                            (acc_mu + value * gau.Mu, acc_sigma2 + value * value * gau.Variance ))
                                                            (-gbias_mu, gbias_var)

                            let marginProb = normcdf(disc_mu / sqrt(disc_sigma2 + betaSquare))
                            if marginProb > maxClassThresh.[label] then maxClassThresh.[label] <- marginProb
                            if marginProb < minClassThresh.[label] then minClassThresh.[label] <- marginProb

                            if label_relevant then
                                record_posterior.[label].[!testInstanceID] <- marginProb
                            else
                                record_posterior.[label].[!testInstanceID] <- -marginProb

                            let threshId = ref 0
                            while !threshId < noThreshGrid && marginProb >= threshGrid.[!threshId] do

                                gamma.[label, !threshId] <- gamma.[label, !threshId] + 1
                                if label_relevant then      // for computing reference to the ground truth
                                    eta.[label, !threshId] <- eta.[label, !threshId] + 1
                                threshId := !threshId + 1

                            threshIdx_writer.Write(byte !threshId)
                            !threshId
                        else
                            int (threshIdx_reader.ReadByte())

                    let sampleY = [| 0..(instance.Features.Length - 1) |] |> Array.fold_left (fun acc nonzeroId ->
                                         let (feaId, feaVal) = instance.Features.[nonzeroId]
                                         acc + weightSample.[label, feaId] * feaVal
                                     )  (-biasSample.[sampleId] + betaSample.[sampleId, label])

                    if sampleY > 0.0<_> then
                        beta.[label] <- beta.[label] + 1
                        if threshMaxIdx > 0 then
                            [| 0 .. (threshMaxIdx - 1) |] |> Array.iter (fun threshIdx ->
                                alpha.[label, threshIdx] <- alpha.[label, threshIdx] + 1)
                )
                
                instanceOpt := testData.getNextInstance ()

            if sampleId > 0 then threshIdx_reader.Close() else threshIdx_writer.Close()
            
            // Now compute this sample's contribution to ExpFs
            [| 0..(noClass-1) |] |> Array.iter (fun label ->
                    [| 0..(noThreshGrid-1) |] |> Array.iter (fun threshId ->
                        let denom = beta.[label] + gamma.[label, threshId]
                        if denom > 0 then 
                            noEffective.[label, threshId] <- noEffective.[label, threshId] + 1
                            ExpFs.[label, threshId] <- ExpFs.[label, threshId] + 
                                                        (float alpha.[label, threshId]) / (float denom)
                        )
                    )

            if sampleId = 0 then
                let num_eff_class = ref 0
                let total_opt_F = ref 0.0
                
                [0 .. (noClass - 1)] |> Seq.iter (fun label ->
                    (record_posterior.[label]) |> Array.sort (fun a b ->                         
                            if abs(a) > abs(b) then 1 
                            elif abs(a) < abs(b) then -1 
                            else 0)

                    printf "0:%A, " (abs(record_posterior.[label].[0]))
                    for i in 1 .. 9 do
                        printf "%d:%A, " i (abs(record_posterior.[label].[Int32.of_float ((float testData.NoInstance) * (float i) / 10.0)]))
                    printfn "10:%A" (abs(record_posterior.[label].[testData.NoInstance - 1]))

                    /// Next show the optimal threshold and its corresponding F-score
                    let numerator = ref (record_posterior.[label] |> Array.sum_by (fun a -> if a > 0.0 then 1 else 0))
                    let denominator = ref (testData.NoInstance + !numerator)

                    let optimal_thresh = ref Double.NegativeInfinity
                    let max_Fscore = ref ((float !numerator) / (float !denominator))

                    for i in 0 .. (testData.NoInstance - 1) do
                        denominator := !denominator - 1
                        if record_posterior.[label].[i] > 0.0 then
                            numerator := !numerator - 1

                        let cur_Fscore = (float !numerator) / (float !denominator)

                        if Double.IsNaN cur_Fscore then // if cur_Fscore is nan at some point, then it can't appear in test data
                            max_Fscore := cur_Fscore
                        if cur_Fscore > !max_Fscore then
                            max_Fscore := cur_Fscore
                            optimal_thresh := abs(record_posterior.[label].[i])

                    if not (Double.IsNaN !max_Fscore) then
                        num_eff_class := !num_eff_class + 1
                        total_opt_F := !total_opt_F + !max_Fscore

                    printfn "optimal F = %g, arg = %g\n========================" (200.0 * !max_Fscore) !optimal_thresh
                )

                printfn "Average optimal Fscore = %g, with %d effective classes" 
                            (200.0 * (float !total_opt_F) / (float !num_eff_class)) !num_eff_class
        )

        printfn "\nstatistics collected, costing %g seconds" 
        #if linux
                    (get_time() - startTime)
        #else            
                    (pro.TotalProcessorTime - startTime).TotalSeconds
                    // (System.DateTime.Now - startTime).TotalSeconds
        #endif
        
        printfn "--------------------------------------------------------------------"

        File.Delete buffer_filename
        let maxExpFs = ref 0.0
        let numEff = ref 0

        [| 0..(noClass-1) |] |> Array.iter (fun label ->
              let (maxVal, argMax) = 
                         [| 0..(noThreshGrid-1) |] |> Array.fold_left (fun (curMax, curArgMax) threshId ->      

                             ExpFs.[label, threshId] <- ExpFs.[label, threshId] / (float noEffective.[label, threshId])

                             if ExpFs.[label, threshId] > curMax then
                                 (ExpFs.[label, threshId], threshGrid.[threshId])
                             else
                                 (curMax, curArgMax)
                             ) (Double.MinValue, -1.0)

              if Double.IsNaN ExpFs.[label, noThreshGrid-1] then
                    printfn "%d special" label
                    [0; 1] |> Seq.iter (fun i -> thresholds.[i, label] <- threshGrid.[0])
                    [2; 3] |> Seq.iter (fun i -> thresholds.[i, label] <- threshGrid.[noThreshGrid-1])
                    [4; 5] |> Seq.iter (fun i -> thresholds.[i, label] <- minClassThresh.[label])
                    [6; 7] |> Seq.iter (fun i -> thresholds.[i, label] <- maxClassThresh.[label])                                    
              else
//                    printfn "%d choose %g" label argMax
                    [0; 2; 4; 6] |> Seq.iter (fun i -> thresholds.[i, label] <- argMax)
                    [1; 3; 5; 7] |> Seq.iter (fun i -> thresholds.[i, label] <- argMax / 2.0)    
                    
                    maxExpFs := !maxExpFs + maxVal
                    numEff := !numEff + 1                                        
             ) 

        printfn "Empirical Fscore: %g based on %d classes" (!maxExpFs * 200.0 / (float !numEff)) !numEff

        // Write to file the empiracal estimation of expected F-score
        if modelParam.writeSampleExpFs then  
                         
            let writer = new StreamWriter(myFileOpenWrite(sprintf "sound_check_%s_%s_sample=%d_prior=%A_noise=%A.dat"
                                                            modelParam.datasetName modelParam.filterFname modelParam.noSample 
                                                            modelParam.biasPriorSigma (betaNoise * 1.0<_>))) 
            threshGrid |> Array.iter (fun value ->
                writer.Write(Float.to_string (value) + "\t"))
            writer.Write("\n")

            [| 0..(noClass-1) |] |> Array.iter (fun label ->
                                        [| 0..(noThreshGrid-1) |] |> Array.iter (fun idx ->
                                            writer.Write( sprintf "%.5g\t" (ExpFs.[label, idx] * 200.0)))
                                        writer.Write("\n")
                                    )
            writer.Close ()


        // Write to file the F-score(l(\theta), l^*)
        if modelParam.writeRefFscore then
        
            let writer = new StreamWriter(myFileOpenWrite(sprintf "ground_truth_%s_%s_sample=%d_prior=%A_noise=%A.dat"
                                                            modelParam.datasetName modelParam.filterFname modelParam.noSample 
                                                            modelParam.biasPriorSigma (betaNoise * 1.0<_>)))  
            threshGrid |> Array.iter (fun value ->
                writer.Write(Float.to_string (value) + "\t"))
            writer.Write("\n")
        
            [| 0..(noClass-1) |] |> Array.iter (fun label ->
                                    [| 0..(noThreshGrid-1) |] |> Array.iter (fun threshIdx ->
                                        let value = (float eta.[label, threshIdx]) / (float (gamma.[label, threshIdx] + rho.[label]))
                                        writer.Write( (sprintf "%.5g\t" (value * 200.0))))
                                    writer.Write("\n")
                                )
            writer.Close ()

        (thresholds, thresholdNames)





    //////////////////////////////////////////////////////////////////////
    /// Allows reading testing data for multiple times
    /// Implementation of the sampling posterior to find the threshold
    /// Returns the best threshold found for each category
    /// 4 heuristics are tried in case the sampled weights never output positive for a certain class, 
    /// Also tried the halving trick
    /// In sum, we have smallestTrial+argmax, smallestTrial+argmax/2, greatestTrial+argmax, greatestTrial+argmax/2 
    ///                 smallestProb+argmax, smallestProb+argmax/2, greatestProb+argmax, greatestProb+argmax/2







    //////////////////////////////////////////////////////////////////////
    /// Allows reading testing data for multiple times
    /// Implementation of the sampling posterior to find the threshold
    /// Returns the best threshold found for each category
    /// 4 heuristics are tried in case the sampled weights never output positive for a certain class, 
    /// Also tried the halving trick
    /// In sum, we have smallestTrial+argmax, smallestTrial+argmax/2, greatestTrial+argmax, greatestTrial+argmax/2 
    ///                 smallestProb+argmax, smallestProb+argmax/2, greatestProb+argmax, greatestProb+argmax/2

    member this.correlationDiagnosis_multipleRead_debug (testData:SeqDataSet<'a>, metaData:MetaData, betaSquare:float<'a^2>,
                                                            varBag:DistributionBag<Gaussian<'a>>, weightVar:int[,], 
                                                            thresholdVar:int, trainIter:int) = 

        let noClass = metaData.NoLabel
        let noFeature = metaData.NoFeature
        
        let threshGrid = [| 0.000001; 0.01; 0.05; 0.1; 0.15; 0.2; 0.25; 0.3; 0.35; 0.4; 0.5; 0.55; 0.6; 0.65; 0.7; 0.75; 0.8; 1.1 |]
        let noThreshGrid = threshGrid.Length
        let noSample = modelParam.noSample
        
        let (weightSeed, biasSeed) = this.generateSeeds()         

        let alpha = Array2.zero_create<int> noClass noThreshGrid
        let beta = Array.zero_create<int> noClass
        let gamma = Array2.zero_create<int> noClass noThreshGrid

        let eta = Array2.zero_create noClass noThreshGrid
        let rho = Array.zero_create noClass

        let ExpFs = Array2.zero_create noClass noThreshGrid
        let noEffective = Array2.zero_create noClass noThreshGrid

        let minClassThresh = Array.zero_create noClass        
        let maxClassThresh = Array.zero_create noClass

        let thresholds = Array2.zero_create 8 noClass
        let thresholdNames = [| @"smallestTrial+argmax"; @"smallestTrial+argmax/2"; @"greatestTrial+argmax"; @"greatestTrial+argmax/2"; 
                                @"smallestProb+argmax"; @"smallestProb+argmax/2"; @"greatestProb+argmax"; @"greatestProb+argmax/2"|]

        printfn "Performing correlation diagnosis."
        
        let resultThreshold = Array.init 8 (fun i -> (Array.zero_create noClass))

        let rng = System.Random(1)
        
            
        let weightSample = Array2.zero_create noClass noFeature

            
        let biasRng = System.Random (biasSeed)  // random number generator for the bias
        let (mu, std) = (varBag.[thresholdVar].Mu, varBag.[thresholdVar].StandardDeviation)
        let biasSample = Array.init noSample (fun i -> 
                mu + std * norminv ((float (biasRng.Next(modelParam.maxRandValue))) / (float modelParam.maxRandValue))
            )

        let betaNoise = sqrt(betaSquare)
        let betaSample = Array2.init noSample noClass (fun sample label -> 
                betaNoise * norminv ((float (biasRng.Next(modelParam.maxRandValue))) / (float modelParam.maxRandValue))
            )


        let rngList = Array2.init noClass noFeature (fun label feaId -> System.Random (weightSeed.[label, feaId]))

        #if linux
        let startTime = get_time() // System.DateTime.Now
        #else
        let pro = Process.GetCurrentProcess ()
        let startTime = pro.TotalProcessorTime // System.DateTime.Now
        #endif
        
        [0..(noSample-1)] |> Seq.iter (fun sampleId ->
        
            printf "sample %d (%d). "             
            #if linux
                    sampleId (Int32.of_float (get_time() - startTime))
            #else
                    sampleId (Int32.of_float (pro.TotalProcessorTime - startTime).TotalSeconds)
                    // (Int32.of_float (System.DateTime.Now - startTime).TotalSeconds)
            #endif
        
            [0..(noClass-1)] |> Seq.iter (fun label ->             
                    [0..(noFeature-1)] |> Seq.iter (fun feaId -> 
                            let gau = varBag.[weightVar.[label, feaId]]
                            weightSample.[label, feaId] <- gau.StandardDeviation 
                                                            * norminv ((float (rngList.[label, feaId].Next(modelParam.maxRandValue))) / (float modelParam.maxRandValue))
                                                            + gau.Mu
                        )
                    beta.[label] <- 0
                    [0..(noThreshGrid-1)] |> Seq.iter (fun threshId -> alpha.[label, threshId] <- 0)
                  )
        
            testData.rewind ()            
            let instanceOpt = ref (testData.getNextInstance ())
            
            while Option.is_some !instanceOpt do
                            
                let instance = Option.get !instanceOpt
                let curActiveLabels = instance.Labels
                let nextId = ref 0
                
                if sampleId = 0 then 
                    curActiveLabels |> Array.iter (fun label -> rho.[label] <- rho.[label] + 1)
                
                [0 .. (noClass - 1)] |> Seq.iter (fun label ->
                    let (disc_mu, disc_sigma2) = instance.Features |> Array.fold_left  (fun (acc_mu, acc_sigma2) (feaId, value) ->
                                                    let gau = varBag.[weightVar.[label, feaId]]
                                                    (acc_mu + value * gau.Mu, acc_sigma2 + value * value * gau.Variance ))
                                                    (-varBag.[thresholdVar].Mu, varBag.[thresholdVar].Variance)

                    let marginProb = normcdf(disc_mu / sqrt(disc_sigma2))
                    if marginProb > maxClassThresh.[label] then maxClassThresh.[label] <- marginProb
                    if marginProb < maxClassThresh.[label] then minClassThresh.[label] <- marginProb

                    let label_relevant = !nextId < instance.NoLabel && curActiveLabels.[!nextId] = label
                    if label_relevant then nextId := !nextId + 1
        
                    let threshFlag = threshGrid |> Array.mapi (fun threshId thresh -> 
                        if marginProb > thresh then
                            if sampleId = 0 then                            
                                gamma.[label, threshId] <- gamma.[label, threshId] + 1

                                // now check with ground truth
                                if label_relevant then
                                    eta.[label, threshId] <- eta.[label, threshId] + 1
                            true
                        else
                            false 
                        )
                            
                    let sampleY = [| 0..(instance.Features.Length - 1) |] |> Array.fold_left (fun acc nonzeroId ->
                                         let (feaId, feaVal) = instance.Features.[nonzeroId]
                                         acc + weightSample.[label, feaId] * feaVal
                                     )  (-biasSample.[sampleId] + betaSample.[sampleId, label])

                    if sampleY > 0.0<_> then
                        beta.[label] <- beta.[label] + 1
                        threshFlag |> Array.iteri (fun threshIdx flag ->
                            if flag then alpha.[label, threshIdx] <- alpha.[label, threshIdx] + 1)
                )
                
                instanceOpt := testData.getNextInstance ()

            // Now compute this sample's contribution to ExpFs
            [| 0..(noClass-1) |] |> Array.iter (fun label ->
                    [| 0..(noThreshGrid-1) |] |> Array.iter (fun threshId ->
                        let denom = beta.[label] + gamma.[label, threshId]
                        if denom > 0 then 
                            noEffective.[label, threshId] <- noEffective.[label, threshId] + 1
                            ExpFs.[label, threshId] <- ExpFs.[label, threshId] + (float alpha.[label, threshId]) / (float denom)
                        )
                    )
        )
        
        printfn "\nstatistics collected, costing %g seconds" 
        #if linux
                        (get_time() - startTime)
        #else
                        (pro.TotalProcessorTime - startTime).TotalSeconds
                        // (System.DateTime.Now - startTime).TotalSeconds
        #endif
        
        printfn "--------------------------------------------------------------------"

        let maxExpFs = Array.zero_create 8
        let numEff = Array.zero_create 8
        
        [| 0..(noClass-1) |] |> Array.iter (fun label ->
              let (maxVal, argMax) = 
                         [| 0..(noThreshGrid-1) |] |> Array.fold_left (fun (curMax, curArgMax) threshId ->      

                             ExpFs.[label, threshId] <- ExpFs.[label, threshId] / (float noEffective.[label, threshId])

                             if ExpFs.[label, threshId] > curMax then
                                 (ExpFs.[label, threshId], threshGrid.[threshId])
                             else
                                 (curMax, curArgMax)
                             ) (Double.MinValue, -1.0)

              if Double.IsNaN ExpFs.[label, noThreshGrid-1] then
//                    printfn "%d special" label
                    [0; 1] |> Seq.iter (fun i -> thresholds.[i, label] <- threshGrid.[0])
                    [2; 3] |> Seq.iter (fun i -> thresholds.[i, label] <- threshGrid.[noThreshGrid-1])
                    [4; 5] |> Seq.iter (fun i -> thresholds.[i, label] <- minClassThresh.[label])
                    [6; 7] |> Seq.iter (fun i -> thresholds.[i, label] <- maxClassThresh.[label])                                    
              else
//                    printfn "%d choose %g" label argMax
                    [0; 2; 4; 6] |> Seq.iter (fun i -> thresholds.[i, label] <- argMax)
                    [1; 3; 5; 7] |> Seq.iter (fun i -> thresholds.[i, label] <- argMax / 2.0) 
                     
                    [|0..7|] |> Array.iter (fun i -> maxExpFs.[i] <- maxExpFs.[i] + maxVal)
                    [|0..7|] |> Array.iter (fun i -> numEff.[i] <- numEff.[i] + 1)
             ) 

        [|0..7|] |> Array.iter (fun i -> maxExpFs.[i] <- maxExpFs.[i] / (float numEff.[i]))

        // Write to file the empiracal estimation of expected F-score
        if modelParam.writeSampleExpFs then  
                         
            let writer = new StreamWriter(myFileOpenWrite(sprintf "sound_check_%s_%s_sample=%d_prior=%A_noise=%A.dat"
                                                            modelParam.datasetName modelParam.filterFname modelParam.noSample 
                                                            modelParam.biasPriorSigma (betaNoise * 1.0<_>))) //
            threshGrid |> Array.iter (fun value ->
                writer.Write(Float.to_string (value) + "\t"))
            writer.Write("\n")

            [| 0..(noClass-1) |] |> Array.iter (fun label ->
                                        [| 0..(noThreshGrid-1) |] |> Array.iter (fun idx ->
                                            writer.Write( sprintf "%.5g\t" (ExpFs.[label, idx] * 200.0)))
                                        writer.Write("\n")
                                    )
            writer.Close ()


        // Write to file the F-score(l(\theta), l^*)
        if modelParam.writeRefFscore then
        
            let writer = new StreamWriter(myFileOpenWrite(sprintf "ground_truth_%s_%s_sample=%d_prior=%A_noise=%A.dat"
                                                            modelParam.datasetName modelParam.filterFname modelParam.noSample 
                                                            modelParam.biasPriorSigma (betaNoise * 1.0<_>)))  
            threshGrid |> Array.iter (fun value ->
                writer.Write(Float.to_string (value) + "\t"))
            writer.Write("\n")
        
            [| 0..(noClass-1) |] |> Array.iter (fun label ->
                                    [| 0..(noThreshGrid-1) |] |> Array.iter (fun threshIdx ->
                                        let value = (float eta.[label, threshIdx]) / (float (gamma.[label, threshIdx] + rho.[label]))
                                        writer.Write( (sprintf "%.5g\t" (value * 200.0))))
                                    writer.Write("\n")
                                )
            writer.Close ()

        (thresholds, maxExpFs, thresholdNames)




    /// Use the given probability threshold to calculate the micro-averaged and macro-averaged F1 score
    /// On the test data
    /// Input 
    ///   1. An array of probability threshold, one scalar for each class
    ///   2. Two common thresholds to be shared by all classes (derived by maximizing microaverage and macroaverage respectively)
    member this.fastTestMicroMacroFscore (thresholds:float[,], thresholdName:array<string>) = 

        let noClass = metaData.NoLabel                
        let noTestThreshold = Array2.length1 thresholds
        
        let (varBag, weightVar, thresholdVar) = getModel ("")
        
        let acc_total_pos = Array2.zero_create noTestThreshold noClass
        let acc_true_pos = Array2.zero_create noTestThreshold noClass
        
        let (gbias_mu, gbias_var, gbias_std) = 
                if modelParam.useGlobalBias 
                    then (varBag.[thresholdVar].Mu, varBag.[thresholdVar].Variance, varBag.[thresholdVar].StandardDeviation) 
                else (0.0<_>, 0.0<_>, 0.0<_>)

        printfn "Predicting on testing data."

        let numTestData = ref 0
        let numRealPos = Array.zero_create noClass        
        
        let rec getCumulant (testdata:SeqDataSet<'a>) =
            match testdata.getNextInstance () with
            | None -> ()
            | Some(instance) -> 
//                instance.print (metaData.LabelIdToName, metaData.FeatureIdToName)
                numTestData := !numTestData + 1
                let curActiveLabels = instance.Labels
                let nextId = ref 0
                [0 .. (noClass - 1)] |> Seq.iter (fun label ->
                    let (disc_mu, disc_sigma2) = instance.Features |> Array.fold_left  (fun (acc_mu, acc_sigma2) (feaId, value) ->   
                                                    let gau = varBag.[weightVar.[label, feaId]]
                                                    (acc_mu + value * gau.Mu, acc_sigma2 + value * value * gau.Variance )
                                                    )  (-gbias_mu, gbias_var)
                    let discriminant = disc_mu / sqrt(disc_sigma2)
//                    printfn "cdf = %A" cdf

                    let label_relevant = !nextId < instance.NoLabel && curActiveLabels.[!nextId] = label
                    if label_relevant then numRealPos.[label] <- numRealPos.[label] + 1; nextId := !nextId + 1

                    [0..(noTestThreshold - 1)] |> Seq.iter (fun curTestThreshold ->
                        if discriminant >= thresholds.[curTestThreshold, label] then
                            acc_total_pos.[curTestThreshold, label] <- acc_total_pos.[curTestThreshold, label] + 1
                            if label_relevant then
                                acc_true_pos.[curTestThreshold, label] <- acc_true_pos.[curTestThreshold, label] + 1   
                    )                         
                )

                curActiveLabels |> Array.iter (fun label -> 
                                   [0..(noTestThreshold - 1)] |> Seq.iter (fun curTestThreshold -> 
                                        acc_total_pos.[curTestThreshold, label] <- acc_total_pos.[curTestThreshold, label] + 1
                                        )
                                    )

                getCumulant testdata

        testData.rewind ()
        getCumulant testData

        ///now display something
        [|0..(noClass-1)|] |> Array.iter (fun label ->
                printfn "Class id = %s, Real pos = %d" metaData.LabelIdToName.[label]  numRealPos.[label]
                if noTestThreshold > 7 then
                    [(noTestThreshold - 8)..(noTestThreshold - 1)] |> Seq.iter (fun threshId ->
                        printfn "Threshold name = %s\tvalue = %f\tF-score = %g\tpos_pred = %d" 
                                thresholdName.[threshId] thresholds.[threshId, label] 
                                (200.0 * (float acc_true_pos.[threshId, label]) / (float acc_total_pos.[threshId, label]))
                                (acc_total_pos.[threshId, label] - numRealPos.[label])
                    )                
                else
                    printfn "Threshold name = %s\tvalue = %f\tF-score = %g\tpos_pred = %d" 
                                thresholdName.[0] thresholds.[0, label] 
                                (200.0 * (float acc_true_pos.[0, label]) / (float acc_total_pos.[0, label]))
                                (acc_total_pos.[0, label] - numRealPos.[label])

                printfn "====================================================="
                    
        )

        [0..(noTestThreshold - 1)] |> Seq.iter (fun curTestThreshold -> 
                let (totalTruePos, totalTotalPos, numEffectiveClass, macroaverage_baby, accuracy_baby) = 
                        [|0..(noClass-1)|] |> Array.fold_left (fun (acc1, acc2, acc3, acc4, acc5) label ->
                                    if acc_total_pos.[curTestThreshold, label] > 0 then
                                        ( acc1 + acc_true_pos.[curTestThreshold, label], 
                                          acc2 + acc_total_pos.[curTestThreshold, label],
                                          acc3 + 1,
                                          acc4 + (float acc_true_pos.[curTestThreshold, label]) / (float acc_total_pos.[curTestThreshold, label]),
                                          acc5 + (float (2 * acc_true_pos.[curTestThreshold, label] - acc_total_pos.[curTestThreshold, label] + !numTestData)) / (float !numTestData)
                                        )
                                    else
                                        (acc1, acc2, acc3, acc4, acc5 + 1.0)
                                    ) (0, 0, 0, 0.0, 0.0)
            
                let macroaverage = macroaverage_baby * 200.0 / (float numEffectiveClass)
                let microaverage = 200.0 * (float totalTruePos) / (float totalTotalPos)
                let accuracy = accuracy_baby * 100.0 / (float noClass)

                printfn "Using %s, numEffectiveClass = %d, Macroaverage Fscore = %g, Microaverage Fscore = %g, accuracy = %g"
                           thresholdName.[curTestThreshold] numEffectiveClass macroaverage microaverage accuracy
            )
            
            
    member this.testOnTrain () =
    
        let noClass = metaData.NoLabel
        let noFeature = metaData.NoFeature

        let (varBag, weightVar, thresholdVar) = getModel ("")

        let (gbias_mu, gbias_var, gbias_std) = 
                if modelParam.useGlobalBias 
                    then (varBag.[thresholdVar].Mu, varBag.[thresholdVar].Variance, varBag.[thresholdVar].StandardDeviation) 
                else (0.0<_>, 0.0<_>, 0.0<_>)

        printfn "===========================================================" 
        printfn "Checking on training data."
                  
        let record_posterior = Array.init noClass (fun i -> Array.init<float> trainData.NoInstance (fun _ -> 0.0))

        trainData.rewind ()
        let instanceOpt = ref (trainData.getNextInstance ())
        let trainInstanceID = ref -1

        while Option.is_some !instanceOpt do

            let instance = Option.get !instanceOpt
            let curActiveLabels = instance.Labels
            let nextId = ref 0
            trainInstanceID := !trainInstanceID + 1

            [0 .. (noClass - 1)] |> Seq.iter (fun label ->

                let label_relevant = !nextId < instance.NoLabel && curActiveLabels.[!nextId] = label
                if label_relevant then nextId := !nextId + 1

                let (disc_mu, disc_sigma2) = instance.Features |> Array.fold_left  (fun (acc_mu, acc_sigma2) (feaId, value) ->
                                               let gau = varBag.[weightVar.[label, feaId]]
                                               (acc_mu + value * gau.Mu, acc_sigma2 + value * value * gau.Variance )
                                               ) (-gbias_mu, gbias_var)

                let posterior = disc_mu / sqrt(disc_sigma2)

                let a = if label_relevant then 1.0 else -1.0
                record_posterior.[label].[!trainInstanceID] <- a * exp(posterior / modelParam.threshScaling)
            )

            instanceOpt := trainData.getNextInstance ()

        GC.Collect()

        let num_eff_class = ref 0
        let total_opt_F = ref 0.0  
        let opt_thresh_list = Array2.zero_create 1 noClass    
        
        [0 .. (noClass - 1)] |> Seq.iter (fun label ->
        
            // sort in decreasing order
            (record_posterior.[label]) |> Array.sort (fun a b ->
                    if abs(a) > abs(b) then -1
                    elif abs(a) < abs(b) then 1
                    else 0)

            /// Next show the optimal threshold and its corresponding F-score
            let numerator = ref (record_posterior.[label] |> Array.sum_by (fun a -> if a > 0.0 then 1 else 0))
            let denominator = ref (trainData.NoInstance + !numerator)
            let optimal_thresh = ref Double.NegativeInfinity
            
            let max_Fscore = ref ((float !numerator) / (float !denominator))

            let i = ref (trainData.NoInstance - 1)
            while !i >= 0 do
                denominator := !denominator - 1
                if record_posterior.[label].[!i] > 0.0 then
                    numerator := !numerator - 1

                if !denominator = 0 then // if cur_Fscore is nan at some point, then it can't appear in test data
                    max_Fscore := Double.NaN
                else
                    let cur_Fscore = (float !numerator) / (float !denominator)
                    if cur_Fscore > !max_Fscore then
                        max_Fscore := cur_Fscore
                        optimal_thresh := 
                            if !i > 0 then 
                                exp(log(abs(record_posterior.[label].[!i])) * 0.99
                                            + log(abs(record_posterior.[label].[!i - 1])) * 0.01)
                            else
                                1.0 + abs(record_posterior.[label].[0])
                i := !i - 1

            optimal_thresh := modelParam.threshScaling * log(!optimal_thresh)
            if not (Double.IsNaN !max_Fscore) then
                num_eff_class := !num_eff_class + 1
                total_opt_F := !total_opt_F + !max_Fscore
            else
                failwith "Unexpected"
            opt_thresh_list.[0, label] <- !optimal_thresh
        )

        printfn "Average optimal Fscore on training data = %g, with %d effective classes" 
                    (200.0 * (float !total_opt_F) / (float !num_eff_class)) !num_eff_class
        printfn "===========================================================" 
        
        opt_thresh_list
        

    /// Dump the data for sound check (against ground truth)
    member this.dumpDiagnoise threshGrid (expFs:double[,]) (eta:int[,]) (gamma:int[,]) (rho:int[]) =
    
        let (noClass, noThreshGrid) = (Array2.length1 threshGrid, Array2.length2 threshGrid)
        let betaNoise = sqrt(modelParam.betaSquare)
        
        // Write to file the empiracal estimation of expected F-score
        if modelParam.writeSampleExpFs then  
                         
            let writer = new StreamWriter(myFileOpenWrite(sprintf "sound_check_%s_%s_sample=%d_prior=%A_noise=%A.dat"
                                                            modelParam.datasetName modelParam.filterFname modelParam.noSample 
                                                            modelParam.biasPriorSigma (betaNoise * 1.0<_>))) 
            
            [| 0..(noClass-1) |] |> Array.iter (fun label ->
                [| 0..(noThreshGrid-1) |] |> Array.iter (fun threshIdx ->
                    writer.Write(Float.to_string (threshGrid.[label, threshIdx]) + "\t"))
                writer.Write("\n")
                
                [| 0..(noThreshGrid-1) |] |> Array.iter (fun threshIdx ->
                    writer.Write( sprintf "%.5g\t" (expFs.[label, threshIdx] * 200.0)))

                writer.Write("\n")
            )
                
            writer.Close ()


        // Write to file the F-score(l(\theta), l^*)
        if modelParam.writeRefFscore then
        
            let writer = new StreamWriter(myFileOpenWrite(sprintf "ground_truth_%s_%s_sample=%d_prior=%A_noise=%A.dat"
                                                            modelParam.datasetName modelParam.filterFname modelParam.noSample 
                                                            modelParam.biasPriorSigma (betaNoise * 1.0<_>)))  
            [| 0..(noClass-1) |] |> Array.iter (fun label ->
                [| 0..(noThreshGrid-1) |] |> Array.iter (fun threshIdx ->
                    writer.Write(Float.to_string (threshGrid.[label, threshIdx]) + "\t"))
                writer.Write("\n")
                
                [| 0..(noThreshGrid-1) |] |> Array.iter (fun threshIdx ->
                                        let value = (float eta.[label, threshIdx]) / (float (gamma.[label, threshIdx] + rho.[label]))
                                        writer.Write( (sprintf "%.5g\t" (value * 200.0))))                                        
                writer.Write("\n")
            )
            writer.Close ()



    member this.sampleFscore (threshGrid:float[,]) =
    
        let noClass = metaData.NoLabel
        let noFeature = metaData.NoFeature
        let noSample = modelParam.noSample
        let noThreshGrid = Array2.length2 threshGrid

        let (varBag, weightVar, thresholdVar) = getModel ("")
        
        let alpha = Array2.zero_create<int> noClass noThreshGrid
        let beta = Array.zero_create<int> noClass
        let gamma = Array2.zero_create<int> noClass noThreshGrid

        let eta = Array2.zero_create noClass noThreshGrid
        let rho = Array.zero_create noClass

        let minClassThresh = Array.create noClass Double.MaxValue
        let maxClassThresh = Array.create noClass Double.MinValue
        let ExpFs = Array2.zero_create noClass noThreshGrid
        let noEffective = Array2.zero_create noClass noThreshGrid

        let rng = System.Random(1)
        
        let (min_rand, max_rand) = (modelParam.sampleBarrier, 1.0 - modelParam.sampleBarrier)
        let get_rand () = 
            min_rand + (max_rand - min_rand) * float(rng.Next(modelParam.maxRandValue)) / (float modelParam.maxRandValue - 1.0)

        let (gbias_mu, gbias_var, gbias_std) = 
                if modelParam.useGlobalBias 
                    then (varBag.[thresholdVar].Mu, varBag.[thresholdVar].Variance, varBag.[thresholdVar].StandardDeviation) 
                else (0.0<_>, 0.0<_>, 0.0<_>)

        let gbiasSample = 
            if modelParam.useGlobalBias then
                Array.init noSample (fun i -> gbias_mu + gbias_std * norminv (get_rand ()))
            else
                Array.zero_create noSample

        let weightSample = Array2.zero_create noClass noFeature

        let betaNoise = sqrt(modelParam.betaSquare)
        let betaSample = Array2.init noSample noClass (fun _ _ -> betaNoise * get_rand () )

        #if linux
        let startTime = get_time() // System.DateTime.Now
        #else
        let pro = Process.GetCurrentProcess ()
        let startTime = pro.TotalProcessorTime // System.DateTime.Now
        #endif

        let startTimeDate = System.DateTime.Now // System.DateTime.Now
        
        if not (Directory.Exists (modelParam.root + @"/Dataset/useless")) then
            Directory.CreateDirectory (modelParam.root + @"/Dataset/useless") |> ignore
        let buffer_filename = modelParam.root + @"/Dataset/useless/threshIdx" + (Int64.to_string startTimeDate.Ticks) + ".dat" 
        
        // start sampling
        let hit_class = Array.zero_create noClass
                
        [0..(noSample-1)] |> Seq.iter (fun sampleId ->
        
            printf "sample %d (%d). " 
            #if linux
                    sampleId (Int32.of_float (get_time () - startTime))
            #else
                    sampleId (Int32.of_float (pro.TotalProcessorTime - startTime).TotalSeconds)
                    // (Int32.of_float (System.DateTime.Now - startTime).TotalSeconds)
            #endif 
            
            [0..(noClass-1)] |> Seq.iter (fun label ->             
                    [0..(noFeature-1)] |> Seq.iter (fun feaId -> 
                            let gau = varBag.[weightVar.[label, feaId]]
                            weightSample.[label, feaId] <- gau.StandardDeviation 
                                                            //* norminv ((float (rng.Next(maxRandValue))) / (float maxRandValue))
                                                            * norminv (get_rand ())
                                                            + gau.Mu
                        )
                    beta.[label] <- 0
                    [0..(noThreshGrid-1)] |> Seq.iter (fun threshId -> alpha.[label, threshId] <- 0)
                  )
            
            let threshIdx_reader = if sampleId > 0 then 
                                        new BinaryReader(File.Open(buffer_filename, FileMode.Open, FileAccess.Read, FileShare.None))
                                   else null

            let threshIdx_writer = if sampleId = 0 then 
                                        new BinaryWriter(File.Open(buffer_filename, FileMode.Create, FileAccess.Write, FileShare.None))
                                   else null
        
            testData.rewind ()            
            let instanceOpt = ref (testData.getNextInstance ())
            let testInstanceID = ref -1
            
            while Option.is_some !instanceOpt do
                            
                let instance = Option.get !instanceOpt
                let curActiveLabels = instance.Labels
                let nextId = ref 0
                testInstanceID := !testInstanceID + 1
                
                if sampleId = 0 then        // for computing reference to the ground truth
                    curActiveLabels |> Array.iter (fun label -> rho.[label] <- rho.[label] + 1)
                
                [0 .. (noClass - 1)] |> Seq.iter (fun label ->
                
                    let label_relevant = !nextId < instance.NoLabel && curActiveLabels.[!nextId] = label
                    if label_relevant then nextId := !nextId + 1

                    let threshMaxIdx = 
                        if sampleId = 0 then 
                            let (disc_mu, disc_sigma2) = instance.Features |> Array.fold_left  (fun (acc_mu, acc_sigma2) (feaId, value) ->
                                                            let gau = varBag.[weightVar.[label, feaId]]
                                                            (acc_mu + value * gau.Mu, acc_sigma2 + value * value * gau.Variance ))
                                                            (-gbias_mu, gbias_var)

                            let marginProb = disc_mu / sqrt(disc_sigma2)
                            if marginProb > maxClassThresh.[label] then maxClassThresh.[label] <- marginProb
                            if marginProb < minClassThresh.[label] then minClassThresh.[label] <- marginProb

//                            if label_relevant then
//                                record_posterior.[label].[!testInstanceID] <- marginProb
//                            else
//                                record_posterior.[label].[!testInstanceID] <- -marginProb

                            let threshId = ref 0

                            while !threshId < noThreshGrid && marginProb >= threshGrid.[label, !threshId] do

                                gamma.[label, !threshId] <- gamma.[label, !threshId] + 1
                                if label_relevant then      // for computing reference to the ground truth
                                    eta.[label, !threshId] <- eta.[label, !threshId] + 1
                                threshId := !threshId + 1

                            threshIdx_writer.Write(byte !threshId)
                            !threshId
                        else
                            int (threshIdx_reader.ReadByte())

                    let sampleY = [| 0..(instance.Features.Length - 1) |] |> Array.fold_left (fun acc nonzeroId ->
                                         let (feaId, feaVal) = instance.Features.[nonzeroId]
                                         acc + weightSample.[label, feaId] * feaVal
                                     )  (-gbiasSample.[sampleId] + betaSample.[sampleId, label])

                    if sampleY > 0.0<_> then
                        if hit_class.[label] = 0 then 
                            printfn "hit class %s with average mean = %A" metaData.LabelIdToName.[label] sampleY
                            hit_class.[label] <- 1
                        beta.[label] <- beta.[label] + 1
                        if threshMaxIdx > 0 then
                            [| 0 .. (threshMaxIdx - 1) |] |> Array.iter (fun threshIdx ->
                                alpha.[label, threshIdx] <- alpha.[label, threshIdx] + 1)
                )
                
                instanceOpt := testData.getNextInstance ()

            if sampleId > 0 then threshIdx_reader.Close() else threshIdx_writer.Close()
            
            // Now compute this sample's contribution to ExpFs
            [| 0..(noClass-1) |] |> Array.iter (fun label ->
                    [| 0..(noThreshGrid-1) |] |> Array.iter (fun threshId ->
                        let denom = beta.[label] + gamma.[label, threshId]
                        if denom > 0 then 
                            noEffective.[label, threshId] <- noEffective.[label, threshId] + 1
                            ExpFs.[label, threshId] <- ExpFs.[label, threshId] + 
                                                        (float alpha.[label, threshId]) / (float denom)
                        )
                    )
        )

        printfn "\nstatistics collected, costing %g seconds" 
        #if linux
                (get_time() - startTime) 
        #else
                (pro.TotalProcessorTime - startTime).TotalSeconds
                // (System.DateTime.Now - startTime).TotalSeconds
        #endif
        
        printfn "--------------------------------------------------------------------"

        File.Delete buffer_filename

        (ExpFs, noEffective, minClassThresh, maxClassThresh, eta, gamma, rho)




    /// Use the sampled weights to pick the threshold in the candidate set
    /// Result is recorded in pickedThresh (#algo * class)
    member this.fillPickedThresh (pickedThresh:float[,]) expFs (noEffective:int[,]) 
                                    (minClassThresh:float[]) (maxClassThresh:float[]) (threshGrid:float[,]) =
    
        let maxExpFs = ref 0.0
        let numEff = ref 0
        let noThreshGrid = Array2.length2 expFs

        [| 0..(metaData.NoLabel - 1) |] |> Array.iter (fun label ->
              let (maxVal, argMax) = 
                         [| 0..(noThreshGrid-1) |] |> Array.fold_left (fun (curMax, curArgMax) threshId ->      

                             expFs.[label, threshId] <- expFs.[label, threshId] / (float noEffective.[label, threshId])

                             if expFs.[label, threshId] > curMax then
                                 (expFs.[label, threshId], threshGrid.[label, threshId])
                             else
                                 (curMax, curArgMax)
                             ) (Double.MinValue, -1.0)

              if Double.IsNaN expFs.[label, noThreshGrid - 1] then    // never seen in the samples
                    printfn "%s is never actived by weight samples" metaData.LabelIdToName.[label]
                    [0; 1] |> Seq.iter (fun i -> pickedThresh.[i, label] <- threshGrid.[label, 0])
                    [2; 3] |> Seq.iter (fun i -> pickedThresh.[i, label] <- threshGrid.[label, noThreshGrid-1])
                    [4; 5] |> Seq.iter (fun i -> pickedThresh.[i, label] <- minClassThresh.[label])
                    [6; 7] |> Seq.iter (fun i -> pickedThresh.[i, label] <- maxClassThresh.[label])
                    [8; 9] |> Seq.iter (fun i -> pickedThresh.[i, label] <- maxClassThresh.[label] + 1.0)
              else
//                    printfn "%d choose %g" label argMax
                    [0; 2; 4; 6; 8] |> Seq.iter (fun i -> pickedThresh.[i, label] <- argMax)
                    [1; 3; 5; 7; 9] |> Seq.iter (fun i -> pickedThresh.[i, label] <- norminv(normcdf(argMax) / 2.0))

                    maxExpFs := !maxExpFs + maxVal
                    numEff := !numEff + 1                                        
             ) 

        printfn "Empirical Fscore: %g based on %d classes.  Final thresholds picked." 
                        (!maxExpFs * 200.0 / (float !numEff)) !numEff


    ///////////////////////////////////////////////
    /// Set the threshold candidates
    /// threshGrid must be increasingly sorted
    member this.setThreshold (pickedThreshold:float[,]) (record_posterior:array<array<float>>) 
                                (max_Fscore_list:float[]) (optimal_thresh_list:float[]) (num_test_pos_list:int[]) = 

        
        let num_stretch = modelParam.noThreshGridStretch
        let noThreshGrid = num_stretch * 2 + 2
        let threshGrid = Array2.zero_create metaData.NoLabel noThreshGrid
        
        let noSampledPost = Array.length record_posterior.[0]

        [0 .. (metaData.NoLabel - 1)] |> Seq.iter (fun label ->
        
            let ratio = (float trainData.numPos.[label]) / (float trainData.NoInstance)
            let index =  Int32.of_float ((float noSampledPost) * ratio)                
            threshGrid.[label, num_stretch] <- modelParam.threshScaling * log(abs(record_posterior.[label].[index]))

            let step_multi = 2.0

            let cur_step = ref ((modelParam.threshScaling * log(abs(record_posterior.[label].[0]))
                                    - threshGrid.[label, num_stretch])
                                    * (step_multi - 1.0) / (step_multi ** (float num_stretch) - 1.0))
            for i in 1 .. num_stretch do
                 threshGrid.[label, num_stretch + i] <- threshGrid.[label, num_stretch + i - 1] + !cur_step
                 cur_step := !cur_step * step_multi

            let cur_step = ref ((threshGrid.[label, num_stretch] - 
                                        modelParam.threshScaling * log(abs(record_posterior.[label].[noSampledPost - 1])))
                                    * (step_multi - 1.0) / (step_multi ** (float num_stretch) - 1.0))

            for i in 1 .. num_stretch do
                 threshGrid.[label, num_stretch - i] <- threshGrid.[label, num_stretch - i + 1] - !cur_step
                 cur_step := !cur_step * step_multi

            threshGrid.[label, num_stretch * 2 + 1] <- 1.0 + modelParam.threshScaling * 
                                                                log(abs(record_posterior.[label].[0]))

            pickedThreshold.[10, label] <- threshGrid.[label, num_stretch] 

            printfn "Class id = %s" metaData.LabelIdToName.[label]
            printf "0:%A, " (modelParam.threshScaling * log(abs(record_posterior.[label].[noSampledPost - 1])))
            for i in 1 .. 9 do
                printf "%d:%A, " i (modelParam.threshScaling * 
                                    log(abs(record_posterior.[label].[Int32.of_float ((float noSampledPost) * (float (10 - i)) / 10.0)])))
            printfn "10:%A" (modelParam.threshScaling * log(abs(record_posterior.[label].[0])))
            printfn "picked threshold candidates:"
            for i in 0 .. (num_stretch * 2 + 1) do
                printf "%d:%g\t" (i+1) threshGrid.[label, i]
        
            printfn ""
            if this.allPostSampled () then
                printfn "optimal F = %g, arg = %g\nReal num_pos = %d\n========================" 
                        max_Fscore_list.[label] optimal_thresh_list.[label] num_test_pos_list.[label]
        )

        GC.Collect()
        printfn "Threshold candidates picked"

        threshGrid
        
        
    member this.allPostSampled () =
        testData.NoInstance * metaData.NoLabel < modelParam.maxPostSpace



    ///  Pick threshold candidates by using CMN on test data with heap
    member this.fastSetThreshold (pickedThreshold:float[,]) =
    
        GC.Collect ()
        
        let (varBag, weightVar, thresholdVar) = getModel ("")
        let (gbias_mu, gbias_var, gbias_std) = 
                if modelParam.useGlobalBias 
                    then (varBag.[thresholdVar].Mu, varBag.[thresholdVar].Variance, varBag.[thresholdVar].StandardDeviation) 
                else (0.0<_>, 0.0<_>, 0.0<_>)
 
        let heapArray = Array.init metaData.NoLabel (fun label ->
                let ratio = (float trainData.numPos.[label]) / (float trainData.NoInstance)
                new MinHeap (Int32.of_float(ratio * (float testData.NoInstance)))
            )
            
        printfn "Space allocated for heaps."
            
        let minPosterior = Array.create metaData.NoLabel Double.PositiveInfinity
        let maxPosterior = Array.create metaData.NoLabel Double.NegativeInfinity
//        let check = Array.init metaData.NoLabel (fun _ -> Array.create testData.NoInstance 0.0)

        testData.rewind ()

        let instanceOpt = ref (testData.getNextInstance ())

        let writer_post = 
            if debugInfo.saveTestPosterior then
                new StreamWriter(myFileOpenWrite("post_" + (modelParam.getIDString ()) + ".dat"))
            else
                null
        
        while Option.is_some !instanceOpt do

            let instance = Option.get !instanceOpt

            [0 .. (metaData.NoLabel - 1)] |> Seq.iter (fun label ->

                let (disc_mu, disc_sigma2) = instance.Features |> Array.fold_left  (fun (acc_mu, acc_sigma2) (feaId, value) ->
                                               let gau = varBag.[weightVar.[label, feaId]]
                                               (acc_mu + value * gau.Mu, acc_sigma2 + value * value * gau.Variance )
                                               ) (-gbias_mu, gbias_var)

                let posterior = disc_mu / sqrt(disc_sigma2)
                if debugInfo.saveTestPosterior then 
                    writer_post.Write((Float.to_string posterior) + " ")

                heapArray.[label].add (posterior)
//                check.[label].[testData.CurInstance - 1] <- posterior
                if posterior > maxPosterior.[label] then maxPosterior.[label] <- posterior
                if posterior < minPosterior.[label] then minPosterior.[label] <- posterior
            )

            if debugInfo.saveTestPosterior then writer_post.Write("\n")
            
            instanceOpt := testData.getNextInstance ()

        
//        ///// now check whether the heap is working correctly
//        [0 .. (metaData.NoLabel - 1)] |> Seq.iter (fun label -> 
//                check.[label] |> Array.sort (fun a b -> if a > b then -1 elif a < b then 1 else 0)
//                let ratio = (float trainData.numPos.[label]) / (float trainData.NoInstance)
//                printfn "class id = %s, brutal = %A, by heap = %A" 
//                        metaData.LabelIdToName.[label]
//                        check.[label].[Int32.of_float(ratio * (float testData.NoInstance))]
//                        (heapArray.[label].getTop ())
//            )
        
                
        
        let num_stretch = modelParam.noThreshGridStretch
        let noThreshGrid = num_stretch * 2 + 2
        let threshGrid = Array2.zero_create metaData.NoLabel noThreshGrid
        
        [0 .. (metaData.NoLabel - 1)] |> Seq.iter (fun label ->
        
            threshGrid.[label, num_stretch] <- heapArray.[label].getTop ()

            let step_multi = 2.0

            let cur_step = ref ((maxPosterior.[label] - threshGrid.[label, num_stretch])
                                    * (step_multi - 1.0) / (step_multi ** (float num_stretch) - 1.0))
            for i in 1 .. num_stretch do
                 threshGrid.[label, num_stretch + i] <- threshGrid.[label, num_stretch + i - 1] + !cur_step
                 cur_step := !cur_step * step_multi

            let cur_step = ref ((threshGrid.[label, num_stretch] - minPosterior.[label])
                                    * (step_multi - 1.0) / (step_multi ** (float num_stretch) - 1.0))

            for i in 1 .. num_stretch do
                 threshGrid.[label, num_stretch - i] <- threshGrid.[label, num_stretch - i + 1] - !cur_step
                 cur_step := !cur_step * step_multi

            threshGrid.[label, num_stretch * 2 + 1] <- 1.0 + maxPosterior.[label]

            ////////////////////////
            /// Note this, set the threshold of CMN
            pickedThreshold.[10, label] <- threshGrid.[label, num_stretch] 

            printfn "Class id = %s, picked threshold candidates:" metaData.LabelIdToName.[label]
            for i in 0 .. (num_stretch * 2 + 1) do
                printf "%d:%g\t" (i+1) threshGrid.[label, i]
        
            printfn ""
        )
        
        GC.Collect()
        printfn "Threshold candidates picked"

        threshGrid



    /// Compute the posterior for all (example, class) pair
    /// If space is not enough (taking more than maxPostSpace), the random sample a subset of the test data
    member this.computePosterior () =
    
        GC.Collect ()
        
        let (varBag, weightVar, thresholdVar) = getModel ("")
        let (gbias_mu, gbias_var, gbias_std) = 
                if modelParam.useGlobalBias 
                    then (varBag.[thresholdVar].Mu, varBag.[thresholdVar].Variance, varBag.[thresholdVar].StandardDeviation) 
                else (0.0<_>, 0.0<_>, 0.0<_>)

        let sampledPosteriorFlag = 
            if testData.NoInstance * metaData.NoLabel < modelParam.maxPostSpace then
                Array.create testData.NoInstance 1
            else
                printfn "Lack of space.  Estimating distribution of posterior via sampling."                
                let rng = System.Random(2)
                let thresh = (float modelParam.maxPostSpace) / (float metaData.NoLabel) / (float testData.NoInstance)
                Array.init testData.NoInstance (fun _ -> 
                    if (float (rng.Next(modelParam.maxRandValue))) / (float modelParam.maxRandValue) < thresh then 1 else 0)

        let noSampledPosterior = sampledPosteriorFlag |> Array.sum
        printfn "Number of sampled posterior = %d" noSampledPosterior

        let record_posterior = Array.init metaData.NoLabel (fun i -> Array.init<float> noSampledPosterior (fun _ -> 0.0))
                
        printfn "Space for record_posterior allocated"

        testData.rewind ()

        let instanceOpt = ref (testData.getNextInstance ())

        let writer_post = 
            if debugInfo.saveTestPosterior then
                new StreamWriter(myFileOpenWrite("post_" + (modelParam.getIDString ()) + ".dat"))
            else
                null

        let sampleIndex = ref 0
        
        while Option.is_some !instanceOpt do

            let instance = Option.get !instanceOpt
            let curActiveLabels = instance.Labels
            let nextId = ref 0

            if debugInfo.saveTestPosterior || sampledPosteriorFlag.[testData.CurInstance - 1] = 1 then
                
                [0 .. (metaData.NoLabel - 1)] |> Seq.iter (fun label ->

                    let label_relevant = !nextId < instance.NoLabel && curActiveLabels.[!nextId] = label
                    if label_relevant then nextId := !nextId + 1

                    let (disc_mu, disc_sigma2) = instance.Features |> Array.fold_left  (fun (acc_mu, acc_sigma2) (feaId, value) ->
                                                   let gau = varBag.[weightVar.[label, feaId]]
                                                   (acc_mu + value * gau.Mu, acc_sigma2 + value * value * gau.Variance )
                                                   ) (-gbias_mu, gbias_var)

                    let posterior = disc_mu / sqrt(disc_sigma2)
                    if debugInfo.saveTestPosterior then 
                        writer_post.Write((Float.to_string posterior) + " ")

                    if sampledPosteriorFlag.[testData.CurInstance - 1] = 1 then
                        let a = if label_relevant then 1.0 else -1.0
                        record_posterior.[label].[!sampleIndex] <- a * exp(posterior / modelParam.threshScaling)
                )

                if debugInfo.saveTestPosterior then writer_post.Write("\n")
                if sampledPosteriorFlag.[testData.CurInstance - 1] = 1 then sampleIndex := !sampleIndex + 1

            instanceOpt := testData.getNextInstance ()

        if debugInfo.saveTestPosterior then writer_post.Close()

        record_posterior


    ///  Pretend that we knew the correct label of the test examples, 
    ///  Find the optimal threshold
    ///  This is for checking whether the learning is well done.  Can't be used in practice.
    member this.findOptimalFscore (record_posterior:array<float>[]) (pickedThresh:float[,]) =

        if not (this.allPostSampled ()) then
            [0 .. (metaData.NoLabel - 1)] |> Seq.iter (fun label -> pickedThresh.[13, label] <- 0.0)
            (null, null, null)
        else        
            let num_eff_class = ref 0
            let total_opt_F = ref 0.0
            
            // Just for delayed (aggregated) printing purposes
            let max_Fscore_list = Array.zero_create<float> metaData.NoLabel
            let optimal_thresh_list = Array.zero_create<float> metaData.NoLabel
            let num_test_pos_list = Array.zero_create<int> metaData.NoLabel
                    
            [0 .. (metaData.NoLabel - 1)] |> Seq.iter (fun label ->
            
                // sort in decreasing order
                (record_posterior.[label]) |> Array.sort (fun a b ->
                        if abs(a) > abs(b) then -1
                        elif abs(a) < abs(b) then 1
                        else 0)

                /// Next show the optimal threshold and its corresponding F-score
                let numerator = ref (record_posterior.[label] |> Array.sum_by (fun a -> if a > 0.0 then 1 else 0))
                let denominator = ref (testData.NoInstance + !numerator)

                let optimal_thresh = ref Double.NegativeInfinity
                let max_Fscore = ref ((float !numerator) / (float !denominator))

                let num_pos = ref 0
                let i = ref (testData.NoInstance - 1)
                while !i >= 0 do
                    denominator := !denominator - 1
                    if record_posterior.[label].[!i] > 0.0 then
                        numerator := !numerator - 1
                        num_pos := !num_pos + 1
                    
                    if !denominator = 0 then // if cur_Fscore is nan at some point, then it can't appear in test data
                        max_Fscore := Double.NaN
                    else
                        let cur_Fscore = (float !numerator) / (float !denominator)
                        if cur_Fscore > !max_Fscore then
                            max_Fscore := cur_Fscore
                            optimal_thresh := 
                                if !i > 0 then 
                                    (abs(record_posterior.[label].[!i])
                                                + abs(record_posterior.[label].[!i - 1])) / 2.0
                                else
                                    1.0 + abs(record_posterior.[label].[0])
                    i := !i - 1

                optimal_thresh := modelParam.threshScaling * log(!optimal_thresh) 
                if not (Double.IsNaN !max_Fscore) then
                    num_eff_class := !num_eff_class + 1
                    total_opt_F := !total_opt_F + !max_Fscore
                else
                    optimal_thresh := Double.PositiveInfinity

                pickedThresh.[13, label] <- !optimal_thresh
                max_Fscore_list.[label] <- 200.0 * !max_Fscore
                optimal_thresh_list.[label] <- !optimal_thresh
                num_test_pos_list.[label] <- !num_pos
            )

            printfn "Average optimal Fscore = %g, with %d effective classes" 
                        (200.0 * (float !total_opt_F) / (float !num_eff_class)) !num_eff_class
            
            (max_Fscore_list, optimal_thresh_list, num_test_pos_list)




    
    member this.correlationDiagnosis_multipleRead_CMN trainIter = 

        let noClass = metaData.NoLabel
        let noFeature = metaData.NoFeature

        let (varBag, weightVar, thresholdVar) = getModel ("")


        ///// final result
        let pickedThresh = Array2.zero_create 14 noClass  // stores threshold before normcdf
        [0..(noClass-1)] |> Seq.iter (fun i -> pickedThresh.[11, i] <- 0.0)
        let thresholdNames = [| @"smallestTrial+argmax"; @"smallestTrial+argmax/2"; 
                                @"greatestTrial+argmax"; @"greatestTrial+argmax/2"; 
                                @"smallestProb+argmax"; @"smallestProb+argmax/2"; 
                                @"greatestProb+argmax"; @"greatestProb+argmax/2"; 
                                @"swichOff+argmax"; @"swichOff+argmax/2"; 
                                @"CMN"; @"0.5"; 
                                @"optTrain"; @"oracle" |]

        printfn "Performing correlation diagnosis.  Starting to allocate space for diagnosis."
        
        let threshGrid = 
            if not modelParam.useFastThreshSet then
                let record_posterior = this.computePosterior ()

                let (max_Fscore_list, optimal_thresh_list, num_test_pos_list) = this.findOptimalFscore record_posterior pickedThresh
                
                let result = this.setThreshold pickedThresh record_posterior max_Fscore_list optimal_thresh_list num_test_pos_list
                let record_posterior = null
                result
            else
                this.fastSetThreshold pickedThresh
                
        GC.Collect()
        
        let (ExpFs, noEffective, minClassThresh, maxClassThresh, eta, gamma, rho) = this.sampleFscore threshGrid
        
        this.fillPickedThresh pickedThresh ExpFs noEffective minClassThresh maxClassThresh threshGrid
        
        this.dumpDiagnoise threshGrid ExpFs eta gamma rho

        let trainThresh = this.testOnTrain ()
        [| 0..(metaData.NoLabel - 1) |] |> Array.iter (fun i -> pickedThresh.[12, i] <- trainThresh.[0, i])

        (pickedThresh, thresholdNames)


        

    //////////////////////////////////////////////////////////////////////
    //////////////////////////////////////////////////////////////////////
    
    
    /// Just test on one instance (used to see whether BOMC immediately fits the current training instance).  
    /// For debug only.  NO longer in use.
    member this.testCurrentTrainingExample (datapoint:DataPointSparse<'a>) =
            
            let (varBag, weightVar, thresholdVar) = getModel ("")

            let features = datapoint.Features
            let labels = datapoint.Labels
            let mutable nextLabelIndex = ref 0

            // first compute the correct label

            for curClass in 0 .. (metaData.NoLabel - 1) do
                let discriminant = datapoint.Features |> Array.fold_left  (fun acc (feaId, value) ->
                                            acc + value * varBag.[weightVar.[curClass, feaId]].Mu ) 0.0<_>
                if discriminant > varBag.[thresholdVar].Mu then
                    if !nextLabelIndex < (Array.length labels) && labels.[!nextLabelIndex] = curClass then
                        nextLabelIndex := !nextLabelIndex + 1
                    else
                        failwith (sprintf "oops, current training data misclassified with discriminant %A" discriminant)
                else
                    if !nextLabelIndex < (Array.length labels) && labels.[!nextLabelIndex] = curClass then
                        failwith (sprintf "oops, current training data misclassified with discriminant %A" discriminant)


    member this.getBillbard (numTrain:int, numBill:int) =

        let rng = System.Random(2)

        let result = Array.init numTrain (fun i -> 1 + Int32.of_float((float (numBill * rng.Next(modelParam.maxRandValue))) 
                                                    / (float modelParam.maxRandValue)))
                                            
        for i in 1 .. numBill do
            let cnt = result |> Array.sum_by (fun x -> if x = i then 1 else 0)
            printfn "bill = %d has %d times" i cnt

        result

    ///////////////////////////////////////////////
    // GDF    
    ///////////////////////////////////////////////

    member this.updateEP (instance:DataPointSparse<'a>) =

       let features = instance.Features
       let noActiveFeature = features.Length

       let workspace = Option.get workspace_ep
       let discriminantFactors = workspace.discriminantFactors
       let noisyDiscriminantFactors = workspace.noisyDiscriminantFactors
       let differenceFactors = workspace.differenceFactors
       let lossFactors = workspace.lossFactors

       let varBag = workspace.varBag
       let weightVar = workspace.weightVar
       let thresholdVar = workspace.thresholdVar

       let runningVarBag = workspace.runningVarBag
       let runningFactorList = workspace.runningFactorList       
       let GF = workspace.GF

       for label = 0 to (metaData.NoLabel - 1) do

           // 1st, Build the variable and factor for the discriminant: f_i^c = x_i * w^c
           let discriminantVar = runningVarBag.NewDistribution ()

           let base_idx = label * metaData.NoFeature // instance.NoFea
           let varIdx = Seq.to_array (discriminantVar :: List.init noActiveFeature (fun idx -> base_idx + fst features.[idx])) 
           
           let labelRelevant = instance.ContainsLabel(label)
           let scalingFactor =  1.0
//                               if labelRelevant then 
//                                    sqrt((float (trainData.numSelData - metaData.numPos.[label])) / (float metaData.numPos.[label]))
//                               else 1.0 / log((float (trainData.numSelData - metaData.numPos.[label])) / (float metaData.numPos.[label]))
//                               else sqrt((float metaData.numPos.[label]) / (float (trainData.numSelData - metaData.numPos.[label])))

           discriminantFactors.[label] <- GF.GaussianFastSmallMultiSparseWeightedSumFactor (Array.append [|(0, -1.0)|] features)  
                                           varIdx
                                           (Array.append [|runningVarBag|] (Array.create noActiveFeature varBag)) 
                                           (if printVarName then sprintf "discFact_%d" label else "")
                                           scalingFactor
                                           |> runningFactorList.AddFactor

           // 2nd Incorporate noise by building noise factor and noisy discriminante variable node
           let noisyDiscriminantVar = runningVarBag.NewDistribution ()

           noisyDiscriminantFactors.[label] <-
                   GF.GaussianLikelihoodFactor modelParam.betaSquare (noisyDiscriminantVar, runningVarBag) (discriminantVar, runningVarBag) 
                           (if printVarName then sprintf "noisyDiscFact_%d" label else "")
                           |> runningFactorList.AddFactor

           // 3rd, build difference factor to compare with the threshold
           let posEpsilon = modelParam.epsilon * log((1.0 + float (trainData.NoInstance - trainData.numPos.[label]))
                                                     / (float trainData.numPos.[label]))
//           let posEpsilon = modelParam.epsilon
//           let posEpsilon = modelParam.epsilon * sqrt((float (trainData.NoInstance - metaData.numPos.[label]))
//                                                        / (float metaData.numPos.[label]))         

//           let negEpsilon = modelParam.epsilon / sqrt((float (trainData.NoInstance - metaData.numPos.[label]))
//                                                        / (float metaData.numPos.[label]))
//           let posEpsilon = modelParam.epsilon

           let negEpsilon = modelParam.epsilon

           if modelParam.useGlobalBias then     // use global bias
               let differenceVar = runningVarBag.NewDistribution ()

               differenceFactors.[label] <- 
                       let factorName = if printVarName then sprintf "diffFact_%d" label else ""

                       GF.GaussianWeightedSumFactor 1.0 (-1.0) (differenceVar, runningVarBag) 
                                       (noisyDiscriminantVar, runningVarBag) (thresholdVar, varBag) factorName
                       |> runningFactorList.AddFactor

               lossFactors.[label] <- 
                       if modelParam.labelNoiseRho < 0.0 then
                           if instance.ContainsLabel(label) then
                               GF.GaussianGreaterThanFactor posEpsilon (differenceVar, runningVarBag) 
                                   (if printVarName then sprintf "GreaterFact_%d" label else "")
                                   |> runningFactorList.AddFactor
                           else
                               GF.GaussianLessThanFactor negEpsilon (differenceVar, runningVarBag) 
                                   (if printVarName then sprintf "LessThanFact_%d" label else "")
                                   |> runningFactorList.AddFactor
                       else
                           printfn "label noise rho = %A" modelParam.labelNoiseRho
                           if instance.ContainsLabel(label) then
                               GF.GaussianNoisyFactor posEpsilon modelParam.labelNoiseRho 
                                                        (differenceVar, runningVarBag) 
                                   (if printVarName then sprintf "NoisyGreaterFact_%d" label else "")
                                   |> runningFactorList.AddFactor
                           else
                               GF.GaussianNoisyFactor negEpsilon (1.0 - modelParam.labelNoiseRho) 
                                                        (differenceVar, runningVarBag) 
                                   (if printVarName then sprintf "NoisyLessThanFact_%d" label else "")
                                   |> runningFactorList.AddFactor

           else    // not use global bias

               lossFactors.[label] <- 
                    if modelParam.labelNoiseRho < 0.0 then
                       if instance.ContainsLabel(label) then
                           GF.GaussianGreaterThanFactor posEpsilon (noisyDiscriminantVar, runningVarBag) 
                               (if printVarName then sprintf "GreaterFact_%d" label else "")
                               |> runningFactorList.AddFactor
                       else
                           GF.GaussianLessThanFactor negEpsilon (noisyDiscriminantVar, runningVarBag) 
                               (if printVarName then sprintf "LessThanFact_%d" label else "")
                               |> runningFactorList.AddFactor
                    else
                       if instance.ContainsLabel(label) then
                           GF.GaussianNoisyFactor posEpsilon modelParam.labelNoiseRho 
                               (noisyDiscriminantVar, runningVarBag) 
                               (if printVarName then sprintf "NoisyGreaterFact_%d" label else "")
                               |> runningFactorList.AddFactor
                       else
                           GF.GaussianNoisyFactor negEpsilon (1.0 - modelParam.labelNoiseRho) (noisyDiscriminantVar, runningVarBag) 
                               (if printVarName then sprintf "NoisyLessThanFact_%d" label else "")
                               |> runningFactorList.AddFactor


       ////////////////////////////////////////////////////////////////////
       //  Next schedule message passing
       ////////////////////////////////////////////////////////////////////
       
       // 1st, propagate through the discriminant factor down
       discriminantFactors |> Array.iter (fun factor -> factor.UpdateMessage 0 |> ignore )

       // 2nd, propagate through the noise factor
       noisyDiscriminantFactors |> Array.iter (fun factor -> factor.UpdateMessage 0 |> ignore )
                                               
       // 3rd, loopy EP.
       let mutable curDelta = modelParam.delta * 2.0 + 1.0<_>
       let mutable noIteration = 0

       while (modelParam.useGlobalBias && curDelta > modelParam.delta && noIteration < 20) || 
               (not modelParam.useGlobalBias && noIteration < 1) do

           noIteration <- noIteration + 1
           if debug then printfn "=====    EP = %d starts  ======" noIteration

           curDelta <- if modelParam.useGlobalBias then
                           [| 

                           // Difference factors propagate downwards
                           (differenceFactors |> Array.map (fun factor -> factor.UpdateMessage 0) |> Array.max) ;

                           // GreaterTahn factors propagate upwards
    //                     (greaterThanFactors |> Array.map (fun factor -> factor.UpdateMessage 0) |> Array.max);
                           (lossFactors |> Array.map (fun factor -> factor.UpdateMessage 0) |> Array.max);

                           // Difference factors send message to threshold variable node
                           (differenceFactors |> Array.map (fun factor -> factor.UpdateMessage 2) |> Array.max) 

                           |] |> Array.max
                        else
                           [| 

                           // GreaterTahn factors propagate upwards
                           (lossFactors |> Array.map (fun factor -> factor.UpdateMessage 0) |> Array.max);

                           |] |> Array.max
           if debug then
               printfn "=====    EP iteration %d ends with delta = %A   ======" noIteration curDelta

       workspace.EPIter <- workspace.EPIter + noIteration

       // now put back the posterior
       if modelParam.useGlobalBias then
           // After convergence, propagate back up.  Only needs to propagate from discriminant factor up
           differenceFactors |> Array.iter (fun factor -> factor.UpdateMessage 1 |> ignore)

       // propagate up through the noise factor
       noisyDiscriminantFactors |> Array.iter (fun factor -> factor.UpdateMessage 1 |> ignore)

       discriminantFactors |> Array.iter (fun factor ->
                           for idx in 1 .. noActiveFeature do
                               factor.UpdateMessage idx |> ignore 
                           )
       
       
       runningVarBag.Reset ()
       runningFactorList.Reset ()
       GF.ClearMessages()

//     if (iter + 1) % param.evidenceIter = 0 then
//         printfn "Average evidence = %g" (1.0 / (float param.evidenceIter) * Array.sum evidenceList)
//         printf "details of %d-th iter: " (iter + 1)
//         detail_evidence |> Array.iteri (fun i value -> printf "%d:%g " i value)
//         printfn ""


    member this.testModel iter =

         let (varBag, weightVar, thresholdVar) = getModel ("")

         if modelParam.useValid then
             let (probThreshold, commonThreshold_micro, commonThreshold_macro) = 
                     this.tuneThreshold ()
             this.testMicroMacroFscore(probThreshold, commonThreshold_micro, commonThreshold_macro)

         if modelParam.showEnum then
             printfn "show Enumerate"
             this.enumMicroMacroFscore (testData, metaData, varBag, weightVar, thresholdVar)

         if modelParam.corDiag then
//             let (thresholds, thresholdNames) = this.cmd_threshold (testData, metaData, 
//                                                             betaSquare, varBag, weightVar, thresholdVar, 
//                                                             iter, param, num_active, !num_train_instance)
//             this.fastTestMicroMacroFscore(testData, metaData, thresholds, thresholdNames, 
//                                             varBag, weightVar, thresholdVar, param)
             let (thresholds, thresholdNames) = this.correlationDiagnosis_multipleRead_CMN iter
             this.fastTestMicroMacroFscore(thresholds, thresholdNames)


    member this.updateClosed (instance:DataPointSparse<'a>) = 
    
        if modelParam.useGlobalBias then failwith "Closed form does not allow global threshold"

        let (varBag, weightVar, _) = getModel ("closed")

        let curActiveLabels = instance.Labels
        let nextId = ref 0

        [0 .. (metaData.NoLabel - 1)] |> Seq.iter (fun label ->

//             let posEpsilon = modelParam.epsilon * (1.0 + log((float (trainData.NoInstance - metaData.numPos.[label]))
//                                                        / (float metaData.numPos.[label])))
             let posEpsilon = modelParam.epsilon
             let negEpsilon = modelParam.epsilon

             let label_relevant = !nextId < instance.NoLabel && curActiveLabels.[!nextId] = label
             if label_relevant then nextId := !nextId + 1

             let numTimes = 1
//                            if label_relevant then 
//                                (trainData.NoInstance - trainData.numPos.[label]) / trainData.numPos.[label] 
//                            else 1

             for i = 1 to numTimes do
                 let (disc_mu, disc_sigma2) = instance.Features |> Array.fold_left  (fun (acc_mu, acc_sigma2) (feaId, value) ->
                                                let gau = varBag.[weightVar.[label, feaId]]
                                                (acc_mu + value * gau.Mu, acc_sigma2 + value * value * gau.Variance ))
                                                         (0.0<_>, modelParam.betaSquare)

                 let t = if label_relevant then 
                              (disc_mu - posEpsilon) / sqrt(disc_sigma2) 
                         else (- disc_mu - negEpsilon) / sqrt(disc_sigma2) 

                 let vt = v t 0.0
                 let wt = w t 0.0

                 instance.Features |> Array.iter (fun (feaId, value) ->
                        let gau = varBag.[weightVar.[label, feaId]]
                        let effect_value = if label_relevant then value else -value
                        let effect_value_squre = effect_value * effect_value
                        let new_precisionmean = (gau.Pm + effect_value * vt / sqrt(disc_sigma2)) / 
                                                    (1.0 - effect_value_squre * wt / (gau.Pre * disc_sigma2))
                        let new_precision = gau.Pre * gau.Pre / (gau.Pre - effect_value_squre * wt / disc_sigma2)
                        varBag.[weightVar.[label, feaId]] <- new Gaussian<'a> (new_precisionmean, new_precision)
                 )
        )             

    member this.get_discrepancy () = 
        let max_diff_pm = ref -1.0<_>
        let max_diff_pre = ref -1.0<_>
        
        let (varBag_ep, weightVar_ep, thresholdVar_ep) = getModel ("ep")
        let (varBag_closed, weightVar_closed, thresholdVar_closed) = getModel ("closed")

        for label in 0 .. (metaData.NoLabel - 1) do
            for feaId in 0 .. (metaData.NoFeature - 1) do
            
                let gau1 = varBag_ep.[weightVar_ep.[label, feaId]]
                let gau2 = varBag_closed.[weightVar_closed.[label, feaId]]
                
                if abs(gau1.Pm - gau2.Pm) > !max_diff_pm then
                    max_diff_pm := abs(gau1.Pm - gau2.Pm)
                if abs(gau1.Pre - gau2.Pre) > !max_diff_pre then
                    max_diff_pre := abs(gau1.Pre - gau2.Pre)

        printfn "max_diff_pm = %A,\tmax_diff_pre = %A" !max_diff_pm !max_diff_pre
        
        
    /////////////////////////////////////////////////////
    /// Main function of training
    /////////////////////////////////////////////////////
    member this.Train () =
       
        let evidenceList = Array.zero_create modelParam.evidenceIter
        
        GC.Collect ()

        #if linux
        let startTime = get_time() // System.DateTime.Now
        #else
        let pro = Process.GetCurrentProcess ()
        let startTime = pro.TotalProcessorTime // System.DateTime.Now
        #endif

        let totalTestTime = ref (startTime - startTime)  // (new System.TimeSpan (0L))

        let totalNoFeature = ref 0

        printfn "Data Read.\nstarting time: %A" System.DateTime.Now

        let noShowResult = ref 0        // how many times have we shown the result
        let noInstanceAfterLastDisp = ref 0     // how many instances have passed since last display of result
        let noTimeProcessed = ref 0
        
        let bbSelector = this.getBillbard (trainData.NoInstance, modelParam.numBill)

//        let num_active = Array.zero_create metaData.NoLabel
//        let num_train_instance = ref 0
                
        let rec innerTrain (iter:int) (billID:int) (sweepID:int) =

            match trainData.getNextInstance () with
            | None -> 
                    if billID < modelParam.numBill then
                        trainData.rewind ()
                        printfn "\nRewinding training data.  Starting billboard = %d, training sweep = %d, time elapsed = %f" 
                        #if linux
                            (billID + 1) sweepID (get_time() - startTime)
                        #else
                            (billID + 1) sweepID (pro.TotalProcessorTime - startTime).TotalSeconds
                        #endif
                        innerTrain iter (billID + 1) sweepID
                    elif sweepID < modelParam.numTrSweep then
                        trainData.rewind ()
                        printfn "\nRewinding training data.  Starting billboard 1, training sweep %d" (sweepID + 1)
                        innerTrain iter 1 (sweepID + 1)
                    else
                        ()

            | Some(instance) -> 

              if iter % 1000 = 0 then  GC.Collect()

              if bbSelector.[iter % trainData.NoInstance] = billID then
              
                if (!noTimeProcessed + 1) % 10000 = 0 then
                    printfn "\n%d training examples processed, time elapsed = %g sec" 
                    #if linux
                            (!noTimeProcessed + 1)  (get_time () - startTime)
                    #else
                            (!noTimeProcessed + 1)  (pro.TotalProcessorTime - startTime).TotalSeconds
                            //*%*/ (float (System.DateTime.Now - startTime).TotalSeconds)
                    #endif
                    
                noInstanceAfterLastDisp := !noInstanceAfterLastDisp + 1
                noTimeProcessed := !noTimeProcessed + 1

                let features = instance.Features
                let noActiveFeature = features.Length
                printf "."

                totalNoFeature := !totalNoFeature + noActiveFeature

                if modelParam.useEP then this.updateEP instance
                if modelParam.useClosed then this.updateClosed instance

                if modelParam.useEP && modelParam.useClosed && (iter + 1) % 1 = 0 then this.get_discrepancy ()

                if debugInfo.showWeightEachTrain then
                    if modelParam.useClosed then printfn "Weights by closed:";  (Option.get workspace_closed).showModel () 
                    if modelParam.useEP then printfn "Weights by EP:"; (Option.get workspace_ep).showModel ()
                    Console.ReadLine () |> ignore

                // Now we can clean up all the variables and factors for this data point

                /// Deal with evidence
//                let (evidence, detail_evidence) = runningFactorList.LogNormalisation
//                evidenceList.[iter % param.evidenceIter] <- evidence


                if !noTimeProcessed % modelParam.monitorInterval = 0 then

                    #if linux
                    let startTestTime = get_time()
                    #else
                    let startTestTime = pro.TotalProcessorTime // System.DateTime.Now                    
                    #endif
                    
                    
                    noShowResult := !noShowResult + 1
                    GC.Collect() 

                    printfn "Before calling test, time cost so far = %g." 
                    #if linux
                        (get_time() - startTime)
                    #else
                        (pro.TotalProcessorTime - startTime).TotalSeconds
                            // (System.DateTime.Now - startTime).TotalSeconds
                    #endif
                    
                    if debugInfo.saveFinalModel then saveModel ()

//                    let threshold = this.testOnTrain ()
//                    this.fastTestMicroMacroFscore(threshold, [| "Optimal Train" |])

                    this.testModel iter

                    #if linux
                    totalTestTime := get_time() - startTestTime + !totalTestTime // System.DateTime.Now - startTestTime + !totalTestTime
                    let totalTime = get_time() - startTime  //System.DateTime.Now - startTime
                    #else
                    totalTestTime := pro.TotalProcessorTime - startTestTime + !totalTestTime // System.DateTime.Now - startTestTime + !totalTestTime
                    let totalTime = pro.TotalProcessorTime - startTime //System.DateTime.Now - startTime
                    #endif
                    
                    let totalTrainTime = totalTime - !totalTestTime

                    let denom = float (iter + 1)
                    
                    printfn "Total training iteration %d, training time %f, Averge training time: %f"
                    #if linux
                            (iter+1) totalTrainTime (totalTrainTime / denom)
                    #else
                            (iter+1) totalTrainTime.TotalSeconds ((float totalTrainTime.TotalSeconds) / denom)
                    #endif
                    
                    printfn "Total test time = %f, Average test time: %f, Total Time elapsed: %f"
                    #if linux
                            !totalTestTime  (!totalTestTime / (float !noShowResult))
                            totalTime
                    #else                    
                            (!totalTestTime).TotalSeconds  ((!totalTestTime).TotalSeconds / (float !noShowResult))
                            totalTime.TotalSeconds
                    #endif
                            
                    printfn "Averge EP iter = %f, average #feature = %f" 
                            (if modelParam.useEP then 
                                let temp = (Option.get workspace_ep).EPIter 
                                (Option.get workspace_ep).EPIter <- 0
                                ((float temp) / (float !noInstanceAfterLastDisp)) 
                             else -1.0)
                            ((float !totalNoFeature) / (float !noInstanceAfterLastDisp))

                    totalNoFeature := 0
                    noInstanceAfterLastDisp := 0

              if check_immediate_fit then
                    this.testCurrentTrainingExample instance

              innerTrain (iter+1) billID sweepID

        // Kick off the training
        trainData.rewind ()
        printfn "Starting billboard 1, training sweep 1"
        innerTrain 0 1 1

        printfn "finishing time: %A\nTime cost: %f" 
                #if linux
                    System.DateTime.Now (get_time() - startTime)
                #else
                    System.DateTime.Now ((pro.TotalProcessorTime - startTime).TotalSeconds)
                    // System.DateTime.Now ((System.DateTime.Now - startTime).TotalSeconds)
                #endif
