﻿//--------------------------------------------------------------
// Program.fs           Bayesian Multi-label Classification
//
// 2008 written by Xinhua Zhang
// Microsoft Research Ltd
//--------------------------------------------------------------

#light

open System
open System.IO
open System.Net
open System.Collections.Generic
open MSRC.Inference.Distributions
open MSRC.Inference.Distributions.Gaussians
open MSRC.Inference
open MSRC.Dataset
open MSRC.Graph

open System.Diagnostics

#if linux
open System.Runtime.InteropServices

[<DllImport("get_time.so", EntryPoint="get_time")>]
let get_time() : double = 0.0
#endif


/////////////////////////////////
/// Default values
let ip = ref 3.0
let is = ref 1
let vp = ref 1.0
let vs = ref 1

//let category = ref "topics"

let datasetName = ref "rcv1_topics_3_1"

let vax = ref 1
let trx = ref 10000
let tex = ref 10000

let filterFname = ref ""

let monitorInterval = ref -1
let betaSquare = ref 1.0        // noise
let epsilon = ref 1.0
let delta = ref 0.0001

let useValid = ref false    // whether to use validation data
let useCMN = ref false
let showEnum = ref false
let corDiag = ref true

let weightPriorSigma = ref 1000.0

let threshPriorSigma = ref 1000.0
let useGlobalBias = ref true

let biasPriorSigma = ref 1000.0    
let useBias = ref true
let biasValue = ref 0.01

let featureCutoffValue = ref 1e-20
let normalizeFeatures = ref true

let maxRandValue = ref 100000000
let noSample = ref 10
let noThreshGridStretch = ref 5
let sampleBarrier = ref 0.01
let maxPostSpace = ref  100000000 // 100000 // 100000000
let useFastThreshSet = ref true

let writeSampleExpFs = ref false
let writeRefFscore = ref false

let numTrSweep = ref 1
let numBill = ref 1
let evidenceIter = ref 1000
let labelNoiseRho = ref -1.0

let root = ref @"C:\Research\bomc\trunk"

// ===============================
// Debug related
let showWeightEachTrain = ref false
let showPostEachTrain = ref false
let printTrainData = ref false
let printTestData = ref false
let printLabelIdToName = ref false
let saveFinalModel = ref false
let saveTestPosterior = ref false
let debugDumpDir = ref @"."

let useEP = ref false
let useClosed = ref false


let threshScaling = ref 10.0

let argspec = 
    [   
        ("-root",   Arg.String (fun value -> root := value),        "Root of everything")

        ("-data",   Arg.String (fun value -> datasetName := value), 
                                                                    "Category (must specify if use -data)")

        ("-flt",    Arg.String (fun value -> filterFname := value), 
                                                                    "Filename of filter")
        
        ("-mon",    Arg.Int (fun value -> monitorInterval := value), 
                                                                    "Monitor interval")
        ("-bs",     Arg.Float (fun value -> betaSquare := value),    
                                                                    "Value of Beta Square of noise")
        ("-epsi",   Arg.Float (fun value -> epsilon := value),      "Value of epsilon in GreaterThan factor")
        ("-delta",  Arg.Float (fun value -> delta := value),        "EP convergence criteria")
        
        ("-va",     Arg.Set useValid,                               "Use validation to tune threshold")
        ("-cmn",    Arg.Set useCMN,                                 "Use CMN to tune threshold")
        ("-enum",   Arg.Set showEnum,                               "Enumerate the threshold and see results")        
        ("-cor",    Arg.Set corDiag,                                "Diagnose the correlation of threshold and performance")
        
        ("-nsam",   Arg.Int (fun value -> noSample := value),       "Number of sample for correlation diagnosis")        
        ("-wsam",   Arg.Set writeSampleExpFs,                       "Write to file the ExpFscore by sampling")
        ("-wref",   Arg.Set writeRefFscore,                         "Write to file the reference fscore")

        ("-wsig",   Arg.Float (fun value -> weightPriorSigma := value), "Sigma of weight prior")
        
        ("-bsig",   Arg.Float (fun value -> biasPriorSigma := value
                                            useBias := value > 0.0),
                                                                    "Sigma of local bias prior")
                                                                    
        ("-gsig",   Arg.Float (fun value -> threshPriorSigma := value
                                            useGlobalBias := value > 0.0), 
                                                                    "Sigma of global threshold prior")

        ("-trSw",   Arg.Int (fun value -> numTrSweep := value),     
                                                                    "Number of times to retrain the data")
        ("-bill",   Arg.Int (fun value -> numBill := value),        
                                                                    "Number of billbard for each sweep")
        ("-eiter",  Arg.Int (fun value -> evidenceIter := value),   
                                                                    "Number of lastest iterations for computing evidence")    
        ("-rho",    Arg.Float (fun value -> labelNoiseRho := value),    
                                                                    "level of label confidence")
        
        ("-smodel", Arg.Set saveFinalModel,                         "Save final model to file")
        ("-spost",  Arg.Set saveTestPosterior,                      "Save posterior of the test data")
        
        ("-ep",     Arg.Set useEP,                                  "Use EP to update the posterior")
        ("-closed",   Arg.Set useClosed,                             "Use closed form to update the posterior")
    ]


/// Dump the parameters to the screen
//let display_parameters() =
//    
//    printfn "Dumping parameter settings."
//    printfn "ip = %.3g, is = %d, vp = %.3g, vs = %d, vax = %d, trx = %d, tex = %d" !ip !is !vp !vs !vax !trx !tex
//    printfn "Dataset = %s, root = %s" !datasetName !root 
    
    
/// (filename_train_X, filename_train_Y, filename_valid_X, filename_valid_Y, filename_test_X, filename_test_Y, dataDirectory) =
let constructFilenames (ip, is, vp, vs, category, root) = 
    
    let directory = sprintf "%s/rcv1_%s/idf_%.3g_%d/valid_%.3g_%d" root category ip is vp vs
    let prefix = sprintf "%s/rcv1_%s_%.3g_%d_%.3g_%d" directory category ip is vp vs
                                  
    (   prefix + @"_train_X.dat", prefix + @"_train_Y.dat", 
        prefix + @"_valid_X.dat", prefix + @"_valid_Y.dat", 
        prefix + @"_test_X.dat", prefix + @"_test_Y.dat", 
        directory)


let myFileOpenWrite (filename:string) =
    File.Open(filename, FileMode.Create, FileAccess.Write, FileShare.None);


let my_trial () = 
    
    let writer = new BinaryWriter(myFileOpenWrite("aaa.dat")) //
      
    printfn "%A" System.DateTime.Now  
    let a = ref 1
    while !a < 100000 do
        a := !a + 1      
        writer.Write(byte (!a % 16))
    
    writer.Close ()
    
    printfn "%A" System.DateTime.Now  
    
    let reader = new BinaryReader(File.Open("aaa.dat", FileMode.Open, FileAccess.Read, FileShare.None))
    
    let a = ref 1
    while !a < 50 do
        a := !a + 1
        let content = int (reader.ReadByte())
        printfn "%d" content
    
    reader.Close ()
    
    printfn "done"


do

    #if linux
    printfn "Testing get_time, current time = %f" (get_time())
    #endif
//     let pro = Process.GetCurrentProcess ()
//     for i in 1..10000 do
//         let result = ref 0
//         for c in 1..200000000 do
//             result := !result + c
//         done
//                     
//         let cur_time = pro.TotalProcessorTime
//         printfn "pro.Id = %d, curProdId = %d, result = %d, time = %f" 
//             pro.Id (Process.GetCurrentProcess().Id) !result (get_time())
//         printfn "current time = (%u, %u, %u, %u) %g" 
//             cur_time.Days cur_time.Hours cur_time.Minutes cur_time.Seconds
//             cur_time.TotalSecond            

//     let pro = Process.GetCurrentProcess ()
//     
//     let pre_time = pro.TotalProcessorTime.TotalSeconds
//     let pre_span = pro.TotalProcessorTime
//     printfn "before: %g" pre_time
//     let c = [|1..100000000|] |> Array.fold_left (fun acc x -> acc + x) 0
//     let after_time = pro.TotalProcessorTime.TotalSeconds
//     let after_span = pro.TotalProcessorTime
//     printfn "after: %g" after_time
//     printfn "Difference = %g, span = %A" (after_time - pre_time) (after_span - pre_span)
    
    
    Arg.parse argspec (fun fileName -> ()) "Help."

    /// Some useful information on the screen
    
    
    printfn "Multilabel classifier v 1.0"
    printfn "2009 written by Xinhua Zhang based on TrueSkill"
    printfn "Microsoft Research Ltd."
    printfn "--------------------------------------------------------------------"
        
    printfn "Root = %s" !root    
    printfn "hostname = %s" (System.Net.Dns.GetHostName ())
 
       
    let sortAndSave = false
    let randomizeAndSave = false

    let debugInfo = new DebugParam (!showWeightEachTrain, !showPostEachTrain, 
                                        !printTrainData, !printTestData, !printLabelIdToName,
                                        !saveFinalModel, !saveTestPosterior,
                                        !debugDumpDir)

    let modelParam = new ModelParameter<Weight> (!datasetName, !filterFname, !root,
                                            !monitorInterval,
                                            !betaSquare, !epsilon, !delta,
                                            !useBias, !biasValue, !normalizeFeatures, !featureCutoffValue,
                                            !numBill, !numTrSweep,
                                            !useValid, !useCMN, !showEnum, !corDiag, 
                                            !maxRandValue, !noSample, !noThreshGridStretch, !sampleBarrier, !maxPostSpace, !useFastThreshSet,
                                            !writeSampleExpFs, !writeRefFscore,                                             
                                            !weightPriorSigma, !biasPriorSigma, !threshPriorSigma, !useGlobalBias, 
                                            !labelNoiseRho, !evidenceIter,
                                            !useEP, !useClosed,
                                            !threshScaling)
                                            
    let dataBundle = new DataBundle<Weight> (!root, !datasetName, !filterFname, !normalizeFeatures, debugInfo, modelParam)    
        
    let learningTask = new LearningTask<Weight> (dataBundle, modelParam, debugInfo)

    learningTask.print()

    let learner = new GraphicalModelADF<Weight> (learningTask)  //
        
    /// The big command: train the model and test on test data
//    learner.Train (trainData, validData, testData, metaData, param, false, -1, dataDirectory, !numBill, !numTrSweep)

    learner.Train ()


    // write the data in the format that other toolboxes can read
       
//    if sortAndSave then
//        printfn "Start writing for sortAndSave"
//        trainData.sortAndSaveToFile()
//        validData.sortAndSaveToFile()
//        testData.sortAndSaveToFile()        
//        failwith "want to exit the program, doesn't know how, so use exception"
        
    // Randomize documents.  This is done offline.
//    if randomizeAndSave then
//        printfn "Start writing for randomizeAndSave"
//        let seed = 1
//        trainData.randomizeAndSave(seed)
//        validData.randomizeAndSave(seed)
//        testData.randomizeAndSave(seed)
//        failwith "want to exit the program, doesn't now how, so use exception"
    
    


